from ._answer import AnswerPattern, answer
from ._match import includes, match
from ._metric import (
    CORRECT,
    INCORRECT,
    NOANSWER,
    PARTIAL,
    Metric,
    Score,
    Value,
    ValueToFloat,
    metric,
    value_to_float,
)
from ._metrics.accuracy import accuracy
from ._metrics.mean import mean
from ._metrics.std import bootstrap_std
from ._model import model_graded_fact, model_graded_qa
from ._pattern import pattern
from ._scorer import (
    Scorer,
    Target,
    scorer,
)

__all__ = [
    "includes",
    "match",
    "model_graded_qa",
    "model_graded_fact",
    "answer",
    "pattern",
    "AnswerPattern",
    "Scorer",
    "Target",
    "scorer",
    "accuracy",
    "bootstrap_std",
    "mean",
    "Metric",
    "metric",
    "Score",
    "Value",
    "ValueToFloat",
    "value_to_float",
    "CORRECT",
    "INCORRECT",
    "PARTIAL",
    "NOANSWER",
]



from inspect_ai._util.constants import PKG_NAME
from inspect_ai._util.registry import registry_info, registry_lookup
from inspect_ai.scorer import Metric, Score, metric


def test_registry_namespaces() -> None:
    # define a local metric which we can lookup by simple name
    @metric(name="local_accuracy")
    def accuracy1(correct: str = "C") -> Metric:
        def metric(scores: list[Score]) -> int | float:
            return 1

        return metric

    assert registry_lookup("metric", "local_accuracy")

    # confirm that inspect_ai builtins have their namespace auto-appended
    info = registry_info(registry_lookup("metric", f"{PKG_NAME}/accuracy"))
    assert info
    assert info.name == f"{PKG_NAME}/accuracy"



from typing import Any, Literal, TypedDict

import httpx
from tenacity import (
    RetryError,
    retry,
    retry_if_exception,
    stop_after_attempt,
    stop_after_delay,
    wait_exponential_jitter,
)

from inspect_ai._util.constants import DEFAULT_MAX_RETRIES
from inspect_ai._util.retry import httpx_should_retry, log_retry_attempt

from ._model import (
    ChatMessage,
    GenerateConfig,
)
from ._tool import ToolInfo


async def chat_api_request(
    client: httpx.AsyncClient,
    model_name: str,
    url: str,
    headers: dict[str, Any],
    json: Any,
    config: GenerateConfig,
) -> Any:
    # provide default max_retries
    max_retries = config.max_retries if config.max_retries else DEFAULT_MAX_RETRIES

    # define call w/ retry policy
    @retry(
        wait=wait_exponential_jitter(),
        stop=(
            (stop_after_attempt(max_retries) | stop_after_delay(config.timeout))
            if config.timeout
            else stop_after_attempt(max_retries)
        ),
        retry=retry_if_exception(httpx_should_retry),
        before_sleep=log_retry_attempt(model_name),
    )
    async def call_api() -> Any:
        response = await client.post(url=url, headers=headers, json=json)
        response.raise_for_status()
        return response.json()

    # make the call
    return await call_api()


def chat_api_input(input: list[ChatMessage]) -> list[dict[str, str]]:
    """Prepare chat prompt data for sending in an HTTP POST request.

    Many chat APIs (e.g. Mistral and CloudFlare) take the OpenAI
    role/content data structure. This is a convenience function that
    takes the `input` to `generate()` and converts it into a JSON
    serializable object that conforms to this structure.

    Args:
        input (list[ChatMessage]): Input to generate from

    Returns:
       Dict that conforms to OpenAI role/content data structure.
    """
    return [dict(role=message.role, content=message.text) for message in input]


class ChatApiFunction(TypedDict, total=False):
    name: str
    """The name of the function to be called.

    Must be a-z, A-Z, 0-9, or contain underscores and dashes, with a maximum length
    of 64.
    """

    description: str
    """
    A description of what the function does, used by the model to choose when and
    how to call the function.
    """

    parameters: dict[str, object]
    """The parameters the functions accepts, described as a JSON Schema object.

    See the
    [guide](https://platform.openai.com/docs/guides/text-generation/function-calling)
    for examples, and the
    [JSON Schema reference](https://json-schema.org/understanding-json-schema/) for
    documentation about the format.

    Omitting `parameters` defines a function with an empty parameter list.
    """


class ChatApiTool(TypedDict, total=False):
    """Tool for use the model during generation."""

    type: Literal["function"]
    """Tool type (currently only function is supported)"""

    function: ChatApiFunction
    """Type information for function to be called"""


def chat_api_tool(tool: ToolInfo) -> ChatApiTool:
    """JSON schema definition for a tool to be called by the model.

    Both OpenAI and Mistral use JSON schema for their tool definition
    (others will likely follow suit).

    Args:
       tool (ToolInfo): Tool definition

    Returns:
       Name and JSON schema for tool parameters and return value.
    """
    # build params
    properties: dict[str, Any] = {}
    required: list[str] = []
    for param in tool.params:
        properties[param.name] = dict(
            type=param.type,
            description=param.description,
        )
        if not param.optional:
            required.append(param.name)

    # define tool
    return ChatApiTool(
        type="function",
        function=ChatApiFunction(
            name=tool.name,
            description=tool.description,
            parameters=dict(
                type="object",
                properties=properties,
                required=required,
            ),
        ),
    )


# When calling chat_api_request() we use tenacity as the retry wrapper, so
# checking for rate limit errors needs to punch through the RetryError and
# look at its `__cause__`. we've observed CloudFlare giving transient 500
# status as well as a ReadTimeout, so we count these as rate limit errors
def is_chat_api_rate_limit(ex: BaseException) -> bool:
    return isinstance(ex, RetryError) and (
        (
            isinstance(ex.__cause__, httpx.HTTPStatusError)
            and (
                ex.__cause__.response.status_code == 429
                or ex.__cause__.response.status_code == 500
            )
        )
        or isinstance(ex.__cause__, httpx.ReadTimeout)
    )



import click

from inspect_ai._util.dotenv import init_dotenv

from .. import __version__
from .eval import eval_command
from .info import info_command
from .list import list_command
from .score import score_command
from .view import view_command


@click.group(invoke_without_command=True)
@click.option(
    "--version",
    type=bool,
    is_flag=True,
    default=False,
    help="Print the Inspect version.",
)
@click.pass_context
def inspect(ctx: click.Context, version: bool) -> None:
    # if this was a subcommand then allow it to execute
    if ctx.invoked_subcommand is not None:
        return

    if version:
        print(__version__)
        ctx.exit()
    else:
        click.echo(ctx.get_help())
        ctx.exit()


inspect.add_command(eval_command)
inspect.add_command(score_command)
inspect.add_command(view_command)
inspect.add_command(list_command)
inspect.add_command(info_command)


def main() -> None:
    init_dotenv()
    inspect(auto_envvar_prefix="INSPECT")


if __name__ == "__main__":
    main()



from inspect_ai.model import (
    ChatMessageSystem,
    ToolChoice,
)

from .._solver import Generate, Solver, TaskState, solver
from .._util import append_system_message
from .tool import Tool
from .tool_def import tool_defs


@solver
def use_tools(
    tools: Tool | list[Tool] | None = None, tool_choice: ToolChoice = "auto"
) -> Solver:
    """
    Solver that inject tools into the task state to be used in generate().

    Args:
        tools (Tool | list[Tool]): one or more tools to inject into the task state.
        tool_choice (ToolChoice | None): Directive indicating which
          tools the model should use.

    Returns:
        A solver that injects the tools and tool_choice into the task state.
    """
    # create tool defs
    tools = tools if isinstance(tools, list) else [tools] if tools else None
    tdefs = tool_defs(tools) if tools else None

    async def solve(state: TaskState, generate: Generate) -> TaskState:
        # register the tools
        if tools and tdefs:
            state.tools.extend(tools)

            # append the tools system prompts. mark the 'source' of messages
            # as tool so they can be removed if tool_choice == "none"
            for tool in tdefs:
                if tool.prompt:
                    append_system_message(
                        state.messages,
                        ChatMessageSystem(content=tool.prompt, tool=tool.name),
                    )

        # set tool choice (note you can call this function w/o tools
        # for just the side effect of enabling/disabling tool usage)
        state.tool_choice = tool_choice

        # return state
        return state

    return solve



import pytest
from utils import skip_if_github_action, skip_if_no_openai, skip_if_no_together

from inspect_ai.model import ChatMessageUser, GenerateConfig, ModelOutput, get_model


async def generate_with_logprobs(model_name, **model_kwargs) -> ModelOutput:
    model = get_model(
        model_name,
        config=GenerateConfig(
            logprobs=True, top_logprobs=2, temperature=0.001, max_tokens=50
        ),
        **model_kwargs,
    )

    message = ChatMessageUser(content="Hello.")
    return await model.generate(input=[message])


@pytest.mark.asyncio
@skip_if_no_openai
async def test_openai_logprobs() -> None:
    response = await generate_with_logprobs("openai/gpt-3.5-turbo")
    assert response.choices[0].logprobs is not None
    assert response.choices[0].logprobs.content[0].top_logprobs is not None
    assert len(response.choices[0].logprobs.content[0].top_logprobs) == 2


@pytest.mark.asyncio
@skip_if_no_together
async def test_together_logprobs() -> None:
    response = await generate_with_logprobs("together/lmsys/vicuna-13b-v1.5")
    assert (
        response.choices[0].logprobs is not None
        and response.choices[0].logprobs.content[0].top_logprobs
        is None  # together only ever returns top-1, so top_logprobs should always be None
    )


@pytest.mark.asyncio
@skip_if_github_action
async def test_hf_logprobs() -> None:
    response = await generate_with_logprobs(
        "hf/EleutherAI/pythia-70m",
        chat_template="{% for message in messages %}{{ message.content }}{% endfor %}",
    )
    assert (
        response.choices[0].logprobs
        and response.choices[0].logprobs.content[0].top_logprobs is not None
    )
    assert len(response.choices[0].logprobs.content[0].top_logprobs) == 2



from dataclasses import dataclass
from typing import (
    Any,
    Literal,
    Union,
)

from inspect_ai._util.json import JSONType


@dataclass
class ToolParam:
    name: str
    """Parameter name."""

    type: JSONType
    """JSON type of parameter."""

    description: str
    """Description of parameter."""

    optional: bool
    """Is the parameter optional"""


@dataclass
class ToolInfo:
    name: str
    """Tool name."""

    description: str
    """Tool description."""

    params: list[ToolParam]
    """Tool parameters"""


@dataclass
class ToolCall:
    id: str
    """Unique identifer for tool call."""

    function: str
    """Function called."""

    arguments: dict[str, Any]
    """Arguments to function."""

    type: Literal["function"]
    """Type of tool call (currently only 'function')"""


@dataclass
class ToolFunction:
    name: str
    """The name of the function to call."""


ToolChoice = Union[Literal["none", "auto"], ToolFunction]
"""Specify which tool to call.

"auto" means the model decides; "none" means never call a tool; and
ToolFunction instructs the model to call a specific function.
"""



from pathlib import Path

PKG_AUTHOR = "UK AI Safety Institute"
PKG_AUTHOR_DIR = "UK-AISI"
PKG_NAME = Path(__file__).parent.parent.stem
PKG_PATH = Path(__file__).parent.parent
DEFAULT_EPOCHS = 1
DEFAULT_MAX_RETRIES = 5
DEFAULT_TIMEOUT = 120
DEFAULT_MAX_CONNECTIONS = 10
DEFAULT_MAX_TOKENS = 1024
DEFAULT_VIEW_PORT = 7575
DEFAULT_SERVER_HOST = "127.0.0.1"
HTTP = 15
HTTP_LOG_LEVEL = "HTTP"
DEFAULT_LOG_LEVEL = "warning"
SCORED_SUFFIX = "-scored"



import os

import pytest

from inspect_ai import eval
from inspect_ai.solver import tool


def skip_if_env_var(var: str, exists=True):
    condition = (var in os.environ.keys()) if exists else (var not in os.environ.keys())
    return pytest.mark.skipif(
        condition,
        reason=f"Test doesn't work without {var} environment variable defined.",
    )


def skip_if_no_openai(func):
    return skip_if_env_var("OPENAI_API_KEY", exists=False)(func)


def skip_if_no_anthropic(func):
    return skip_if_env_var("ANTHROPIC_API_KEY", exists=False)(func)


def skip_if_no_google(func):
    return skip_if_env_var("GOOGLE_API_KEY", exists=False)(func)


def skip_if_no_mistral(func):
    return skip_if_env_var("MISTRAL_API_KEY", exists=False)(func)


def skip_if_no_cloudflare(func):
    return skip_if_env_var("CLOUDFLARE_API_TOKEN", exists=False)(func)


def skip_if_no_together(func):
    return skip_if_env_var("TOGETHER_API_KEY", exists=False)(func)


def skip_if_no_azureai(func):
    return skip_if_env_var("AZURE_API_KEY", exists=False)(func)


def skip_if_github_action(func):
    return skip_if_env_var("GITHUB_ACTIONS", exists=True)(func)


def run_example(example: str, model: str):
    example_file = os.path.join("examples", example)
    return eval(example_file, model=model, limit=1)


# define tool
@tool(
    prompt="""If you are given a math problem of any kind,
    please use the addition tool to compute the result.""",
    params={"color": "metadata.color"},
)
def addition():
    async def add(color: str, x: int, y: int):
        """
        Tool for adding two numbers.

        Args:
            color (str): Color
            x (int): First number to add.
            y (int): Second number to add.

        Returns:
            The sum of the two numbers.
        """
        return x + y

    return add



from logging import getLogger
from typing import cast

import numpy as np

from .._metric import (
    Metric,
    Score,
    ValueToFloat,
    metric,
    value_to_float,
)

logger = getLogger(__name__)


@metric
def bootstrap_std(
    num_samples: int = 1000, to_float: ValueToFloat = value_to_float()
) -> Metric:
    """Standard deviation of a bootstrapped estimate of the mean.

    Args:
       num_samples (int): Number of bootstrap samples to take.
       to_float (ValueToFloat): Function for mapping
         Value to float for computing metrics. The default
         `value_to_float()` maps CORRECT ("C") to 1.0,
         INCORRECT ("I") to 0, PARTIAL ("P") to 0.5, and
         NOANSWER ("N") to 0, casts numeric values to
         float directly, and prints a warning and returns
         0 if the Value is a complex object (list or dict).

    Returns:
       bootstrap_std metric
    """

    def metric(scores: list[Score]) -> float:
        values = [to_float(score.value) for score in scores]
        std = np.std(
            [
                np.mean(np.random.choice(values, len(values), replace=True))
                for _ in range(num_samples)
            ]
        )
        return cast(float, std.item())

    return metric



import re
import string


def strip_punctuation(s: str) -> str:
    return s.strip(string.whitespace + string.punctuation)


def strip_numeric_punctuation(s: str) -> str:
    # strip $, €, £, and ,
    stripped = re.sub(r"[$,£,€]", "", s)
    # strip . if it's followed by a space, the end of the string,
    # or a non-digit character
    stripped = re.sub(r"\.(?=\s|$|\D)", "", stripped)
    return stripped



import pytest

from inspect_ai.model import (
    ChatMessageAssistant,
    ChatMessageUser,
    ContentImage,
    ContentText,
)
from inspect_ai.model._model import collapse_consecutive_assistant_messages


@pytest.fixture
def user_message_str():
    return ChatMessageUser(content="User message")


@pytest.fixture
def user_message_image_and_str():
    return ChatMessageUser(
        content=[ContentImage(image="foo"), ContentText(text="Message")]
    )


@pytest.fixture
def assistant_message_str():
    return ChatMessageAssistant(content="Assistant message")


def test_collapse_consecutive_assistant_messages_single_assistant_message(
    assistant_message_str,
):
    messages = [assistant_message_str]
    assert collapse_consecutive_assistant_messages(messages) == messages


def test_collapse_consecutive_assistant_messages_alternating_messages(
    user_message_str, user_message_image_and_str, assistant_message_str
):
    messages = [user_message_str]
    assert collapse_consecutive_assistant_messages(messages) == messages

    messages = [user_message_str, assistant_message_str]
    assert collapse_consecutive_assistant_messages(messages) == messages

    messages = [user_message_str, assistant_message_str, user_message_str]
    assert collapse_consecutive_assistant_messages(messages) == messages

    messages = [user_message_str, assistant_message_str, user_message_image_and_str]
    assert collapse_consecutive_assistant_messages(messages) == messages


def test_collapse_consecutive_assistant_messages_consecutive_assistaht_messages(
    assistant_message_str,
):
    messages = [assistant_message_str, assistant_message_str, assistant_message_str]
    assert len(collapse_consecutive_assistant_messages(messages)) == 1



import importlib.util
import os


def running_in_notebook() -> bool:
    try:
        from IPython import get_ipython  # type: ignore

        if "IPKernelApp" not in get_ipython().config:  # type: ignore
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def platform_init() -> None:
    # if we are running in a notebook, confirm that we have ipywidgets
    if running_in_notebook():
        # check for required packages
        if not have_package("ipywidgets"):
            raise ModuleNotFoundError(
                "To using inspect_ai within a notebook, please install ipywidgets with:\n\n"
                + "pip install ipywidgets\n"
            )

        # activate nest_asyncio (required so we operate properly within
        # the Jupyter async event loop)
        import nest_asyncio  # type: ignore

        nest_asyncio.apply()


def have_package(package: str) -> bool:
    return importlib.util.find_spec(package) is not None


def is_running_in_jupyterlab() -> bool:
    return os.getenv("JPY_SESSION_NAME", None) is not None


def is_running_in_vscode() -> bool:
    # Check if running in VS Code Jupyter notebook or interactive window
    if (
        os.getenv("VSCODE_IPYTHON_KERNEL") is not None
        or os.getenv("VSCODE_CLI_REQUIRE_TOKEN") is not None
        or os.getenv("VSCODE_PID") is not None
        or os.getenv("VSCODE_CWD") is not None
    ):
        return True
    # Check if running in a VS Code terminal
    if os.getenv("TERM_PROGRAM") == "vscode":
        return True

    # If none of the conditions are met, we assume it's not running in VS Code
    return False


def is_windows() -> bool:
    return os.name == "nt"



from datetime import datetime
from typing import Literal


def iso_now(
    timespec: Literal[
        "auto", "hours", "minutes", "seconds", "milliseconds" "microseconds"
    ] = "seconds",
) -> str:
    return datetime.now().isoformat(timespec=timespec)



import os

from .._model import StopReason


def as_stop_reason(reason: str | None) -> StopReason:
    """Encode common reason strings into standard StopReason."""
    match reason:
        case "stop" | "eos":
            return "stop"
        case "length" | "content_filter":
            return reason
        case "model_length":
            return "length"
        case "tool_calls" | "function_call":
            return "tool_calls"
        case _:
            return "unknown"


def model_base_url(base_url: str | None, env_vars: str | list[str]) -> str | None:
    if base_url:
        return base_url

    if isinstance(env_vars, str):
        env_vars = [env_vars]

    for env_var in env_vars:
        base_url = os.getenv(env_var, None)
        if base_url:
            return base_url

    return os.getenv("INSPECT_EVAL_MODEL_BASE_URL", None)



from typing import (
    Any,
    Callable,
    Protocol,
    Sequence,
    TypeVar,
    Union,
    cast,
    overload,
    runtime_checkable,
)

from inspect_ai._util.registry import (
    RegistryInfo,
    registry_add,
    registry_create,
    registry_info,
    registry_name,
    registry_tag,
)
from inspect_ai.solver import TaskState

from ._metric import Metric, Score


class Target(Sequence[str]):
    """Target for scoring.

    Target is a sequence of one or more strings. Use the
    `text` property to access the value as a single string.
    """

    def __init__(self, target: str | list[str]) -> None:
        self.target = target if isinstance(target, list) else [target]

    @overload
    def __getitem__(self, index: int) -> str: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[str]: ...

    def __getitem__(self, index: Union[int, slice]) -> Union[str, Sequence[str]]:
        return self.target[index]

    def __len__(self) -> int:
        return len(self.target)

    @property
    def text(self) -> str:
        return "".join(self.target)


@runtime_checkable
class Scorer(Protocol):
    r"""Score model outputs.

    Evaluate the passed outputs and targets and return a
    dictionary with scoring outcomes and context.

    Args:
        state (TaskState): Task state
        target (Target): Ideal target for the output.
    """

    async def __call__(self, state: TaskState, target: Target) -> Score: ...


ScorerType = TypeVar("ScorerType", Callable[..., Scorer], type[Scorer])
r"""Scorer type.

Valid scorer types include:
 - Functions that return a Scorer
 - Classes derivied from Scorer
"""


def scorer_register(scorer: ScorerType, name: str = "") -> ScorerType:
    r"""Register a function or class as a scorer.

    Args:
        scorer (ScorerType):
            Scorer, function that returns a Scorer, or class
            deriving from the Scorer protocol.
        name (str): Name of scorer (Optional, defaults to object name)

    Returns:
        Scorer with registry attributes.
    """
    scorer_name = name if name else getattr(scorer, "__name__")
    registry_add(scorer, RegistryInfo(type="scorer", name=scorer_name))
    return scorer


def scorer_create(name: str, **kwargs: Any) -> Scorer:
    r"""Create a Scorer based on its registered name.

    Args:
        name (str): Name of scorer (Optional, defaults to object name)
        **kwargs (dict): Optional creation arguments for the scorer

    Returns:
        Scorer with registry info attribute
    """
    return cast(Scorer, registry_create("scorer", name, **kwargs))


def scorer(
    metrics: list[Metric], name: str | None = None, **metadata: Any
) -> Callable[[Callable[..., Scorer]], Callable[..., Scorer]]:
    r"""Decorator for registering scorers.

    Args:
        metrics (list[Metric]): One or more metrics to calculate
            over the scores.
        name (str | None):
            Optional name for scorer. If the decorator has no name
            argument then the name of the underlying ScorerType
            object will be used to automatically assign a name.
        **metadata (dict[str,Any]): Additional values to serialize
            in metadata.

    Returns:
        Scorer with registry attributes.

    """

    def wrapper(scorer_type: ScorerType) -> ScorerType:
        # determine the name (explicit or implicit from object)
        scorer_name = registry_name(
            scorer_type, name if name else getattr(scorer_type, "__name__")
        )

        # wrap instatiations of scorer so they carry registry info and metrics
        def scorer_wrapper(*args: Any, **kwargs: Any) -> Scorer:
            scorer = scorer_type(*args, **kwargs)

            registry_tag(
                scorer_type,
                scorer,
                RegistryInfo(
                    type="scorer",
                    name=scorer_name,
                    metadata={SCORER_METRICS: metrics} | metadata,
                ),
                *args,
                **kwargs,
            )
            return scorer

        # register the scorer
        return scorer_register(cast(ScorerType, scorer_wrapper), scorer_name)

    return wrapper


def scorer_metrics(scorer: Scorer) -> list[Metric]:
    return cast(list[Metric], registry_info(scorer).metadata[SCORER_METRICS])


SCORER_METRICS = "metrics"



import abc
import asyncio
import functools
import os
from contextvars import ContextVar
from copy import deepcopy
from typing import Any, Callable, Literal, Type, Union, cast

from pydantic import BaseModel, Field
from tenacity import (
    retry,
    retry_if_exception,
    stop_after_attempt,
    stop_after_delay,
    stop_never,
    wait_exponential_jitter,
)
from typing_extensions import TypedDict

from inspect_ai._util.constants import (
    DEFAULT_MAX_CONNECTIONS,
    PKG_NAME,
)
from inspect_ai._util.platform import platform_init
from inspect_ai._util.registry import RegistryInfo, registry_find, registry_info
from inspect_ai._util.retry import log_rate_limit_retry
from inspect_ai.util import concurrency
from inspect_ai.util._context.concurrency import using_concurrency

from ._tool import ToolCall, ToolChoice, ToolFunction, ToolInfo


class GenerateConfigArgs(TypedDict, total=False):
    """Type for kwargs that selectively override GenerateConfig."""

    max_retries: int | None
    """Maximum number of times to retry request (defaults to 5)."""

    timeout: int | None
    """Request timeout (in seconds)."""

    max_connections: int | None
    """Maximum number of concurrent connections to Model API (default is model specific)."""

    system_message: str | None
    """Override the default system message."""

    max_tokens: int | None
    """The maximum number of tokens that can be generated in the completion (default is model specific)."""

    top_p: float | None
    """An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass."""

    temperature: float | None
    """What sampling temperature to use, between 0 and 2. Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic."""

    stop_seqs: list[str] | None
    """Sequences where the API will stop generating further tokens. The returned text will not contain the stop sequence."""

    best_of: int | None
    """Generates best_of completions server-side and returns the 'best' (the one with the highest log probability per token). OpenAI only."""

    frequency_penalty: float | None
    """Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim. OpenAI only."""

    presence_penalty: float | None
    """Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics. OpenAI only."""

    logit_bias: dict[int, float] | None
    """Map token Ids to an associated bias value from -100 to 100 (e.g. "42=10,43=-10"). OpenAI only."""

    seed: int | None
    """Random seed. OpenAI only. OpenAI and Mistral only."""

    suffix: str | None
    """The suffix that comes after a completion of inserted text. OpenAI only."""

    top_k: int | None
    """Randomly sample the next word from the top_k most likely next words. Anthropic, Google, and HuggingFace only."""

    num_choices: int | None
    """How many chat completion choices to generate for each input message. Open AI, Google, and TogetherAI only."""

    logprobs: bool | None
    """Return log probabilities of the output tokens. OpenAI, TogetherAI, and Huggingface only."""

    top_logprobs: int | None
    """Number of most likely tokens (0-20) to return at each token position, each with an associated log probability. OpenAI and Huggingface only."""


class GenerateConfig(BaseModel):
    """Base class for model generation configs."""

    max_retries: int | None = Field(default=None)
    """Maximum number of times to retry request (defaults to 5)."""

    timeout: int | None = Field(default=None)
    """Request timeout (in seconds)."""

    max_connections: int | None = Field(default=None)
    """Maximum number of concurrent connections to Model API (default is model specific)."""

    system_message: str | None = Field(default=None)
    """Override the default system message."""

    max_tokens: int | None = Field(default=None)
    """The maximum number of tokens that can be generated in the completion (default is model specific)."""

    top_p: float | None = Field(default=None)
    """An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass."""

    temperature: float | None = Field(default=None)
    """What sampling temperature to use, between 0 and 2. Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic."""

    stop_seqs: list[str] | None = Field(default=None)
    """Sequences where the API will stop generating further tokens. The returned text will not contain the stop sequence."""

    best_of: int | None = Field(default=None)
    """Generates best_of completions server-side and returns the 'best' (the one with the highest log probability per token). OpenAI only."""

    frequency_penalty: float | None = Field(default=None)
    """Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim. OpenAI only."""

    presence_penalty: float | None = Field(default=None)
    """Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics. OpenAI only."""

    logit_bias: dict[int, float] | None = Field(default=None)
    """Map token Ids to an associated bias value from -100 to 100 (e.g. "42=10,43=-10"). OpenAI only."""

    seed: int | None = Field(default=None)
    """Random seed. OpenAI only. OpenAI and Mistral only."""

    suffix: str | None = Field(default=None)
    """The suffix that comes after a completion of inserted text. OpenAI only."""

    top_k: int | None = Field(default=None)
    """Randomly sample the next word from the top_k most likely next words. Anthropic, Google, and HuggingFace only."""

    num_choices: int | None = Field(default=None)
    """How many chat completion choices to generate for each input message. Open AI, Google, and TogetherAI only."""

    logprobs: bool | None = Field(default=None)
    """Return log probabilities of the output tokens. OpenAI, TogetherAI, and Huggingface only."""

    top_logprobs: int | None = Field(default=None)
    """Number of most likely tokens (0-20) to return at each token position, each with an associated log probability. OpenAI and Huggingface only."""

    def merge(
        self, other: Union["GenerateConfig", GenerateConfigArgs]
    ) -> "GenerateConfig":
        """Merge another model configuration into this one.

        Args:
           other (Union[GenerateConfig, GenerateConfigArgs]):
              Configuration to merge.

        Returns:
           Merged configuration.
        """
        if not isinstance(other, GenerateConfig):
            other = GenerateConfig(**other)
        config_keys = list(GenerateConfigArgs.__mutable_keys__)  # type: ignore
        config = deepcopy(self)
        for key in config_keys:
            value = getattr(other, key, None)
            if value is not None:
                setattr(config, key, value)
        return config


class ContentText(BaseModel):
    type: Literal["text"] = Field(default="text")
    """Type."""

    text: str
    """Text content."""


class ContentImage(BaseModel):
    type: Literal["image"] = Field(default="image")
    """Type."""

    image: str
    """Either a URL of the image or the base64 encoded image data."""

    detail: Literal["auto", "low", "high"] = Field(default="auto")
    """Specifies the detail level of the image.

    Currently only supported for OpenAI. Learn more in the
    [Vision guide](https://platform.openai.com/docs/guides/vision/low-or-high-fidelity-image-understanding).
    """


Content = Union[ContentText, ContentImage]
"""Content sent to or received from a model."""


class ChatMessageBase(BaseModel):
    content: str | list[Content]
    """Content (simple string or list of string|image content)"""

    source: Literal["input", "generate"] | None = Field(default=None)
    """Source of message."""

    @property
    def text(self) -> str:
        """Get the text content of this message.

        ChatMessage content is very general and can contain either
        a simple text value or a list of content parts (each of which
        can either be text or an image). Solvers (e.g. for prompt
        engineering) often need to interact with chat messages with
        the assumption that they are a simple string. The text
        property returns either the plain str content, or if the
        content is a list of text and images, the text items
        concatenated together (separated by newline)

        Returns: Text content of `ChatMessage` If this message does
          not have text content then "" is returned.
        """
        if isinstance(self.content, str):
            return self.content
        else:
            all_text = [
                content.text for content in self.content if content.type == "text"
            ]
            return "\n".join(all_text)

    @text.setter
    def text(self, text: str) -> None:
        """Set the primary text content for this message.

        ChatMessage content is very general and can contain either
        a simple text value or a list of content parts (each of which
        can either be text or an image). Solvers (e.g. for prompt
        engineering) often need to interact with chat messages with
        the assumption that they are a simple string. The text property
        sets text either to content directly (if it is a `str`) or to
        the first text content item in the message (inserting one at
        the beginning if necessary). If there are multiple text content
        items in the message then after the set there will be only
        one remaining (image content will remain).
        """
        if isinstance(self.content, str):
            self.content = text
        else:
            all_images = [
                content for content in self.content if content.type == "image"
            ]
            self.content = [ContentText(text=text)] + all_images


class ChatMessageSystem(ChatMessageBase):
    role: Literal["system"] = Field(default="system")
    """Conversation role."""

    tool: str | None = Field(default=None)
    """Tool that injected this message."""


class ChatMessageUser(ChatMessageBase):
    role: Literal["user"] = Field(default="user")
    """Conversation role."""


class ChatMessageAssistant(ChatMessageBase):
    role: Literal["assistant"] = Field(default="assistant")
    """Conversation role."""

    tool_calls: list[ToolCall] | None = Field(default=None)
    """Tool calls made by the model."""


class ChatMessageTool(ChatMessageBase):
    role: Literal["tool"] = Field(default="tool")
    """Conversation role."""

    tool_call_id: str | None = Field(default=None)
    """ID of tool call."""

    tool_error: str | None = Field(default=None)
    """Error calling tool."""


ChatMessage = Union[
    ChatMessageSystem, ChatMessageUser, ChatMessageAssistant, ChatMessageTool
]
"""Message in a chat conversation"""


class ModelUsage(BaseModel):
    input_tokens: int = Field(default=0)
    """Total input tokens used."""

    output_tokens: int = Field(default=0)
    """Total output tokens used."""

    total_tokens: int = Field(default=0)
    """Total tokens used."""


StopReason = Literal["stop", "length", "tool_calls", "content_filter", "unknown"]
"""Reason that the model stopped generating."""


class TopLogprob(BaseModel):
    token: str
    """The top-kth token represented as a string."""

    logprob: float
    """The log probability value of the model for the top-kth token."""

    bytes: list[int]
    """The top-kth token represented as a byte array (a list of integers)."""


class Logprob(BaseModel):
    token: str
    """The predicted token represented as a string."""

    logprob: float
    """The log probability value of the model for the predicted token."""

    bytes: list[int]
    """The predicted token represented as a byte array (a list of integers)."""

    top_logprobs: list[TopLogprob] | None
    """If the `top_logprobs` argument is greater than 0, this will contain an ordered list of the top K most likely tokens and their log probabilities."""


class Logprobs(BaseModel):
    content: list[Logprob]
    """a (num_generated_tokens,) length list containing the individual log probabilities for each generated token."""


class ChatCompletionChoice(BaseModel):
    message: ChatMessageAssistant
    """Assistant message."""

    stop_reason: StopReason = Field(default="unknown")
    """Reason that the model stopped generating."""

    logprobs: Logprobs | None = Field(default=None)
    """Logprobs."""


class ModelOutput(BaseModel):
    model: str = Field(default="")
    """Model used for generation."""

    choices: list[ChatCompletionChoice] = Field(default=[])
    """Completion choices."""

    usage: ModelUsage | None = Field(default=None)
    """Model token usage"""

    error: str | None = Field(default=None)
    """Error message in the case of content moderation refusals."""

    @property
    def completion(self) -> str:
        """Text of first message choice text."""
        if len(self.choices) > 0:
            return self.choices[0].message.text
        else:
            return ""

    @completion.setter
    def completion(self, completion: str) -> None:
        """Set the text of the first message choice.

        Args:
          completion (str): Text for first message.
        """
        if len(self.choices) > 0:
            self.choices[0].message.text = completion
        else:
            self.choices.append(
                ChatCompletionChoice(
                    message=ChatMessageAssistant(content=completion), stop_reason="stop"
                )
            )

    @staticmethod
    def from_content(
        model: str,
        content: str,
        stop_reason: StopReason = "stop",
        error: str | None = None,
    ) -> "ModelOutput":
        """Convenient method to create ModelOutput from simple text content."""
        return ModelOutput(
            model=model,
            choices=[
                ChatCompletionChoice(
                    message=ChatMessageAssistant(content=content, source="generate"),
                    stop_reason=stop_reason,
                )
            ],
            error=error,
        )


class ModelAPI(abc.ABC):
    """Model API provider."""

    def __init__(
        self, model_name: str, base_url: str | None, config: GenerateConfig
    ) -> None:
        """Create a model API provider.

        Args:
           model_name (str): Model name.
           base_url (str | None): Alternate base URL for model.
           config (GenerateConfig): Model configuration.
        """
        self.model_name = model_name
        self.base_url = base_url
        self.config = config

    @abc.abstractmethod
    async def generate(
        self,
        input: list[ChatMessage],
        tools: list[ToolInfo],
        tool_choice: ToolChoice,
        config: GenerateConfig,
    ) -> ModelOutput:
        """Generate output from the model.

        Args:
          input (str | list[ChatMessage]): Chat message
            input (if a `str` is passed it is convereted
            to a `ChatUserMessage`).
          tools (list[ToolInfo]): Tools available for the
            model to call.
          tool_choice (ToolChoice): Directives to the model
            as to which tools to prefer.
          config (GenerateConfig): Model configuration.

        Returns:
           ModelOutput
        """
        ...

    def max_tokens(self) -> int | None:
        """Default max_tokens for this Model API."""
        return None

    def max_connections(self) -> int:
        """Default max_connections for this Model API."""
        return DEFAULT_MAX_CONNECTIONS

    def connection_key(self) -> str:
        """Key that defines the scope for enforcement of max_connections."""
        return "default"

    def is_rate_limit(self, ex: BaseException) -> bool:
        """Check whether an exception should be considered a rate limit error."""
        return False

    def collapse_user_messages(self) -> bool:
        """Should consecutive user messages be collapsed into a single message."""
        return False

    def collapse_assistant_messages(self) -> bool:
        """Should consecutive assistant messages be collapsed into a single message."""
        return False


class Model:
    """Model interface."""

    def __init__(self, api: ModelAPI, config: GenerateConfig) -> None:
        """Create a model.

        Args:
           api (ModelAPI): Model API provider.
           config (GenerateConfig): Model configuration.
        """
        self.api = api
        self.config = config

        # if using the Model API standalone in a notebook this will
        # get hit before score() or eval() so we activate nest_asyncio
        platform_init()

    @property
    def name(self) -> str:
        """Model name."""
        return self.api.model_name

    def __str__(self) -> str:
        return f"{ModelName(self)}"

    async def generate(
        self,
        input: str | list[ChatMessage],
        tools: list[ToolInfo] = [],
        tool_choice: ToolChoice | None = None,
        config: GenerateConfig = GenerateConfig(),
    ) -> ModelOutput:
        """Generate output from the model.

        Args:
          input (str | list[ChatMessage]): Chat message
            input (if a `str` is passed it is convereted
            to a `ChatUserMessage`).
          tools (list[ToolInfo]): Tools available for the
            model to call.
          tool_choice (ToolChoice): Directives to the model
            as to which tools to prefer.
          config (GenerateConfig): Model configuration.

        Returns:
           ModelOutput
        """
        # merge with config from init
        config = self.config.merge(config)

        # provide max_tokens from the model api if required
        config.max_tokens = (
            config.max_tokens if config.max_tokens else self.api.max_tokens()
        )

        # normalize input to chat
        if isinstance(input, str):
            input = [ChatMessageUser(content=input)]

        # insert any system message provided in config
        if config.system_message:
            input.insert(0, ChatMessageSystem(content=config.system_message))

        # see if we have a connection semaphore (we won't if we
        # are running outside of an eval()). this is how we enforce
        # concurrency limits (max_connections) for the model
        if using_concurrency():
            async with self._connection_concurrency(config):
                return await self._generate(input, tools, tool_choice, config)

        # no connection semaphore, just proceed straight ot the call
        else:
            return await self._generate(input, tools, tool_choice, config)

    async def _generate(
        self,
        input: list[ChatMessage],
        tools: list[ToolInfo],
        tool_choice: ToolChoice | None,
        config: GenerateConfig,
    ) -> ModelOutput:
        # default to 'auto' for tool_choice (same as underlying model apis)
        tool_choice = tool_choice if tool_choice else "auto"

        # if we have a specific tool selected then filter out the others
        if isinstance(tool_choice, ToolFunction):
            tools = [tool for tool in tools if tool.name == tool_choice.name]

        # if tool_choice is "none" or if there are no tools then fully purge
        # the tools (as some models (e.g. openai and mistral) get confused
        # if you pass them tool definitions along with tool_choice == "none"
        # (they both 'semi' use the tool by placing the arguments in JSON
        # in their output!)
        if tool_choice == "none" or len(tools) == 0:
            tools = []
            tool_choice = "none"

        # filter out system messages for tools not in play on this pass
        if isinstance(input, list):
            # does this message belong to a tool not active on this pass?
            def is_inactive_tool_system_message(message: ChatMessage) -> bool:
                return (
                    isinstance(message, ChatMessageSystem)
                    and message.tool is not None
                    and (
                        tool_choice == "none"
                        or message.tool not in [tool.name for tool in tools]
                    )
                )

            # filter out inactive tool system messages
            input = [
                message
                for message in input
                if not is_inactive_tool_system_message(message)
            ]

            # optionally collapse *consecutive* messages into one - some apis eg anthropic require this
            if self.api.collapse_user_messages():
                input = collapse_consecutive_user_messages(input)

            if self.api.collapse_assistant_messages():
                input = collapse_consecutive_assistant_messages(input)

        # retry for rate limit errors
        @retry(
            wait=wait_exponential_jitter(jitter=5),
            retry=retry_if_exception(self.api.is_rate_limit),
            stop=(
                (
                    stop_after_delay(config.timeout)
                    | stop_after_attempt(config.max_retries)
                )
                if config.timeout and config.max_retries
                else (
                    stop_after_delay(config.timeout)
                    if config.timeout
                    else (
                        stop_after_attempt(config.max_retries)
                        if config.max_retries
                        else stop_never
                    )
                )
            ),
            before_sleep=functools.partial(log_rate_limit_retry, self.api.model_name),
        )
        async def generate() -> ModelOutput:
            return await self.api.generate(
                input=input,
                tools=tools,
                tool_choice=tool_choice,
                config=config,
            )

        # call the model
        model_output = await generate()

        # record usage
        if model_output.usage:
            record_model_usage(f"{self}", model_output.usage)

        # return results
        return model_output

    # semaphore for model generate requests. these can be shared across
    # instances of Model.  This is so that each distinct model endpoint/account
    # combination shares the semaphore -- i.e. if you had 3 instances
    # of a model class (e.g. attacker model, evaluated model, and grader
    # model) in an eval, they won't each get the full max_connections allocated
    # (which would likely cause the rate limit to be exceeded). conversely if
    # you are using distinct models/endpoints/accounts within an eval you should
    # be able get the full max_connections for each of them. subclasses can
    # override the _connection_key() argument to provide a scope within which
    # to enforce max_connections (e.g. by account/api_key, by endpoint, etc.)

    def _connection_concurrency(self, config: GenerateConfig) -> asyncio.Semaphore:
        """Get the appropiate connection semaphore for this model instance."""
        max_connections = (
            config.max_connections
            if config.max_connections
            else self.api.max_connections()
        )
        model_name = ModelName(self)
        return concurrency(
            name=f"{model_name.api}/{model_name.name}",
            concurrency=max_connections,
            key=f"Model{self.api.connection_key()}",
        )


class ModelName:
    r"""Model name (api and specific model served by the api).

    Can be used for structural pattern matching of models against
    various string specifications of models. Used primarily by
    tasks to allow them to condition their behavior on models or
    model famillies.

    String specifications can be fully specified (e.g. openai/gpt-4),
    partially specified by model name only (e.g. gpt-4) or even
    partially specified by a substring of model name (e.g. gpt).
    """

    def __init__(self, model: str | Model) -> None:
        """Create a ModelName.

        Args:
           model: (str | Model): Model to create name for.
        """
        if isinstance(model, str):
            (api, name) = self._parse_model(model)
            if api is None:
                raise ValueError("API not specified for model name")
            self.api = api
            self.name = name
        else:
            # registry names have a package prefix, strip it off
            name = registry_info(model.api).name
            parts = name.split("/")
            self.api = "/".join(parts[1:]) if len(parts) else name
            self.name = model.name

    def __eq__(self, pattern: object) -> bool:
        if isinstance(pattern, str):
            (api, name) = self._parse_model(pattern)
            if (api and api in self.api) and name in self.name:
                return True
            else:
                return name in self.name
        else:
            return False

    def __str__(self) -> str:
        return f"{self.api}/{self.name}"

    def _parse_model(self, model: str) -> tuple[str | None, str]:
        parts = model.split("/")
        if len(parts) > 1:
            return (parts[0], "/".join(parts[1:]))
        else:
            return (None, model)


def get_model(
    model: str | Model | None = None,
    config: GenerateConfig = GenerateConfig(),
    base_url: str | None = None,
    **model_args: Any,
) -> Model:
    """Get an instance of a model.

    Args:
       model (str | Model | None): Model specification.
         If `Model` is passed it is returned unmodified,
         if `None` is passed then the model currently being
         evaluated is returned (or if there is no evaluation
         then the model referred to by `INSPECT_MODEL_NAME`).
       config (GenerationConfig): Configuration for model.
       base_url (str | None): Optional. Alternate base URL for model.
       **model_args (dict[str,Any]): Additional args to
         pass to model constructor.

    Returns:
        Model instance.

    """
    # if the model is None then use the current model from our async
    # context, else try to use INSPECT_EVAL_MODEL (or the legacy INSPECT_MODEL_NAME)
    model = (
        model
        or active_model()
        or os.getenv("INSPECT_EVAL_MODEL", None)
        or os.getenv("INSPECT_MODEL_NAME", None)
    )
    if model is None:
        raise ValueError("No model specified (and no INSPECT_EVAL_MODEL defined)")

    # reflect back model -- we take model as a convenience so that
    # function that accept str | Model can always call get_model and
    # have it resolve correctly (even if trivially)
    if isinstance(model, Model):
        return model

    # split model into api name and model name if necessary
    api_name = None
    parts = model.split("/")
    if len(parts) > 1:
        api_name = parts[0]
        model = "/".join(parts[1:])

    # predicate to match model
    def match_model(info: RegistryInfo) -> bool:
        # strip package name (we use the 'api' as the namespace, we will
        # introduce package scoping if it proves necessary)
        if info.type == "modelapi":
            # model patterns for this provider
            models = info.metadata.get("models", [])

            # if there is an api_name explicitly specified that
            # matches the registered api then trust the model name
            # TODO: this is ugly, we need to clarify the relationship
            # and registraiton semantics of pkg -> provider -> model
            if (
                info.name == api_name
                or info.name.replace(f"{PKG_NAME}/", "") == api_name
            ):
                return True
            # otherwise check for a name match
            else:
                return len([name for name in models if name in model]) > 0
        else:
            return False

    # find a matching model type
    model_types = registry_find(match_model)
    if len(model_types) > 0:
        modelapi_type = cast(type[ModelAPI], model_types[0])
        modelapi_instance = modelapi_type(
            model_name=model, base_url=base_url, config=config, **model_args
        )
        return Model(modelapi_instance, config)

    else:
        from_api = f" from {api_name}" if api_name else ""
        raise ValueError(f"Model name {model}{from_api} not recognized.")


def simple_input_messages(
    input: list[ChatMessage],
    fold_system_message: Callable[[str, str], str] | None = None,
) -> list[ChatMessage]:
    """Transform input messages into a format compatible with more simplistic chat APIs.

    Collects up system messages and folds them into the first user message
    (according to a passed in folding function). Also collapses consecutive
    user messages (as many LLMs require an alternating structure)
    """
    # start by making a deep copy so our mutations don't propagate (e.g. end up in log)
    input = deepcopy(input)

    # aggregate system message from all system messages
    system_message = " ".join(
        [message.text for message in input if isinstance(message, ChatMessageSystem)]
    ).strip()

    # collect all non-system messages and collapse consecutive user messages
    messages: list[ChatMessage] = collapse_consecutive_user_messages(
        [message for message in input if not isinstance(message, ChatMessageSystem)]
    )

    # fold the system message into the first user message
    first_user_message = next(
        message for message in messages if isinstance(message, ChatMessageUser)
    )
    if fold_system_message:
        first_user_message.text = fold_system_message(
            first_user_message.text, system_message
        )
    else:
        first_user_message.text = f"{system_message}\n\n{first_user_message.text}"

    # all done!
    return messages


# Functions to reduce consecutive user messages to a single user message -> required for some models
def collapse_consecutive_user_messages(
    messages: list[ChatMessage],
) -> list[ChatMessage]:
    return functools.reduce(user_message_reducer, messages, [])


# Functions to reduce consecutive assistant messages to a single user message -> required for some models
def collapse_consecutive_assistant_messages(
    messages: list[ChatMessage],
) -> list[ChatMessage]:
    return functools.reduce(assistant_message_reducer, messages, [])


def user_message_reducer(
    messages: list[ChatMessage],
    message: ChatMessage,
) -> list[ChatMessage]:
    return consecutive_message_reducer(messages, message, ChatMessageUser)


def assistant_message_reducer(
    messages: list[ChatMessage],
    message: ChatMessage,
) -> list[ChatMessage]:
    return consecutive_message_reducer(messages, message, ChatMessageAssistant)


def consecutive_message_reducer(
    messages: list[ChatMessage],
    message: ChatMessage,
    message_type: Type[ChatMessage],
) -> list[ChatMessage]:
    if (
        isinstance(message, message_type)
        and len(messages) > 0
        and isinstance(messages[-1], message_type)
    ):
        messages[-1] = combine_messages(messages[-1], message, message_type)
    else:
        messages.append(message)
    return messages


def combine_messages(
    a: ChatMessage, b: ChatMessage, message_type: Type[ChatMessage]
) -> ChatMessageUser:
    if isinstance(a.content, str) and isinstance(b.content, str):
        return message_type(content=f"{a.content}\n{b.content}")
    elif isinstance(a.content, list) and isinstance(b.content, list):
        return message_type(content=a.content + b.content)
    elif isinstance(a.content, str) and isinstance(b.content, list):
        return message_type(content=b.content + [ContentText(text=a.content)])
    else:
        content: list[Content] = [ContentText(text=a.text)]
        content.extend(cast(list[Content], b.content))
        return message_type(content=content)


def init_async_context_model(model: Model) -> None:
    active_model_context_var.set(model)
    init_model_usage()


def active_model() -> Model | None:
    """The model currently being evaluated.

    Returns:
       The model currently being evaluated.
    """
    return active_model_context_var.get(None)


# shared contexts for asyncio tasks
active_model_context_var: ContextVar[Model] = ContextVar("active_model")


def init_model_usage() -> None:
    model_usage_context_var.set({})


def record_model_usage(model: str, usage: ModelUsage) -> None:
    model_usage = model_usage_context_var.get(None)
    if model_usage is not None:
        total_usage = model_usage.get(model, None)
        if not total_usage:
            total_usage = ModelUsage()
        total_usage.input_tokens += usage.input_tokens
        total_usage.output_tokens += usage.output_tokens
        total_usage.total_tokens += usage.total_tokens
        model_usage[model] = total_usage


def collect_model_usage() -> dict[str, ModelUsage]:
    usage = model_usage_context_var.get()
    model_usage_context_var.set({})
    return usage


model_usage_context_var: ContextVar[dict[str, ModelUsage]] = ContextVar("model_usage")



from copy import copy
from typing import Any, cast

from google.ai.generativelanguage import (
    Blob,
    Candidate,
    FunctionCall,
    FunctionResponse,
    Part,
)
from google.api_core.exceptions import TooManyRequests
from google.api_core.retry.retry_base import if_transient_error
from google.generativeai import (  # type: ignore
    GenerationConfig,
    GenerativeModel,
    configure,
)
from google.generativeai.types import (  # type: ignore
    AsyncGenerateContentResponse,
    ContentDict,
    ContentsType,
    FunctionDeclaration,
    HarmBlockThreshold,
    HarmCategory,
    PartDict,
    Tool,
)
from google.protobuf.json_format import ParseDict
from google.protobuf.struct_pb2 import Struct
from typing_extensions import override

from inspect_ai._util.error import exception_message
from inspect_ai._util.images import image_as_data
from inspect_ai.model._providers.util import model_base_url

from .._model import (
    ChatCompletionChoice,
    ChatMessage,
    ChatMessageAssistant,
    ChatMessageSystem,
    ChatMessageTool,
    ChatMessageUser,
    Content,
    ContentImage,
    ContentText,
    GenerateConfig,
    ModelAPI,
    ModelOutput,
    StopReason,
)
from .._tool import ToolCall, ToolChoice, ToolInfo
from .._util import chat_api_tool

VERTEX_SAFETY_SETTINGS = {
    HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: HarmBlockThreshold.BLOCK_NONE,
    HarmCategory.HARM_CATEGORY_HATE_SPEECH: HarmBlockThreshold.BLOCK_NONE,
    HarmCategory.HARM_CATEGORY_HARASSMENT: HarmBlockThreshold.BLOCK_NONE,
    HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_NONE,
}


class GoogleAPI(ModelAPI):
    def __init__(
        self,
        model_name: str,
        base_url: str | None,
        config: GenerateConfig = GenerateConfig(),
        **model_args: Any,
    ) -> None:
        super().__init__(model_name=model_name, base_url=base_url, config=config)

        # configure genai client
        base_url = model_base_url(base_url, "GOOGLE_BASE_URL")
        configure(
            client_options=dict(api_endpoint=base_url),
            **model_args,
        )

        # create model
        self.model = GenerativeModel(self.model_name)

    async def generate(
        self,
        input: list[ChatMessage],
        tools: list[ToolInfo],
        tool_choice: ToolChoice,
        config: GenerateConfig,
    ) -> ModelOutput:
        parameters = GenerationConfig(
            candidate_count=config.num_choices,
            temperature=config.temperature,
            top_p=config.top_p,
            top_k=config.top_k,
            max_output_tokens=config.max_tokens,
            stop_sequences=config.stop_seqs,
        )

        try:
            # google-native messages
            messages = await as_chat_messages(input)

            # cast to AsyncGenerateContentResponse since we passed stream=False
            response = cast(
                AsyncGenerateContentResponse,
                await self.model.generate_content_async(
                    contents=messages,
                    safety_settings=VERTEX_SAFETY_SETTINGS,
                    generation_config=parameters,
                    tools=chat_tools(tools) if len(tools) > 0 else None,
                    stream=False,
                ),
            )
            choices = completion_choices_from_candidates(response.candidates)
            choice = choices[0]
            return ModelOutput(model=self.model_name, choices=[choice])
        except ValueError as ex:
            # If a safety filter is triggered, the response will be empty and a ValueError will be raised
            return ModelOutput.from_content(
                self.model_name,
                "Sorry, but I can't assist with that",
                "content_filter",
                exception_message(ex),
            )

    @override
    def is_rate_limit(self, ex: BaseException) -> bool:
        return isinstance(ex, TooManyRequests)

    @override
    def connection_key(self) -> str:
        """Scope for enforcing max_connections (could also use endpoint)."""
        return self.model_name


async def as_chat_messages(messages: list[ChatMessage]) -> list[ContentsType]:
    # google does not support system messages so filter them out to start with
    system_messages = [message for message in messages if message.role == "system"]
    supported_messages = [message for message in messages if message.role != "system"]

    # build google chat messages
    chat_messages = [await content_dict(message) for message in supported_messages]

    # we want the system messages to be prepended to the first user message
    # (if there is no first user message then prepend one)
    prepend_system_messages(chat_messages, system_messages)

    # return messages
    return chat_messages


async def content_dict(
    message: ChatMessageUser | ChatMessageAssistant | ChatMessageTool,
) -> ContentDict:
    if isinstance(message, ChatMessageUser):
        return ContentDict(
            role="user",
            parts=(
                [PartDict(text=message.content)]
                if isinstance(message.content, str)
                else [await content_part(content) for content in message.content]
            ),
        )
    elif isinstance(message, ChatMessageAssistant):
        if message.tool_calls is not None:
            content_parts = [
                Part(
                    function_call=FunctionCall(
                        name=tool_call.function,
                        args=ParseDict(js_dict=tool_call.arguments, message=Struct()),
                    )
                )
                for tool_call in message.tool_calls
            ]
            if message.content:
                content_parts.append(Part(text=message.content))
            return ContentDict(role="model", parts=content_parts)
        else:
            return ContentDict(role="model", parts=[Part(text=message.content)])
    elif isinstance(message, ChatMessageTool):
        response = FunctionResponse(
            name=message.tool_call_id,
            response=ParseDict(
                js_dict={
                    "content": (
                        message.tool_error
                        if message.tool_error is not None
                        else message.content
                    )
                },
                message=Struct(),
            ),
        )
        return ContentDict(role="function", parts=[Part(function_response=response)])


async def content_part(content: Content | str) -> PartDict:
    if isinstance(content, str):
        return PartDict(text=content)
    elif isinstance(content, ContentText):
        return PartDict(text=content.text)
    else:
        return PartDict(inline_data=await chat_content_image_to_blob(content))


async def chat_content_image_to_blob(image: ContentImage) -> Blob:
    image_url = image.image
    image_bytes, mime_type = await image_as_data(image_url)
    return Blob(mime_type=mime_type, data=image_bytes)


def prepend_system_messages(
    messages: list[ContentDict], system_messages: list[ChatMessageSystem]
) -> None:
    # create system_parts
    system_parts = [Part(text=message.content) for message in system_messages]

    # we want the system messages to be prepended to the first user message
    # (if there is no first user message then prepend one)
    if messages[0].get("role") == "user":
        messages[0]["parts"] = system_parts + messages[0].get("parts", [])
    else:
        messages.insert(0, ContentDict(role="user", parts=system_parts))


def chat_tools(tools: list[ToolInfo]) -> list[Tool]:
    chat_tools = [chat_api_tool(tool) for tool in tools]
    declarations = [
        FunctionDeclaration(
            name=tool["function"]["name"],
            description=tool["function"]["description"],
            parameters=tool["function"]["parameters"],
        )
        for tool in chat_tools
    ]
    return [Tool(declarations)]


def completion_choice_from_candidate(candidate: Candidate) -> ChatCompletionChoice:
    # check for completion text
    content = " ".join(
        [part.text for part in candidate.content.parts if part.text is not None]
    )

    # now tool calls
    tool_calls: list[ToolCall] = []
    for part in candidate.content.parts:
        if part.function_call:
            arguments: dict[str, Any] = {}
            for key in part.function_call.args:
                val = part.function_call.args[key]
                arguments[key] = val
            tool_calls.append(
                ToolCall(
                    type="function",
                    id=part.function_call.name,
                    function=part.function_call.name,
                    arguments=arguments,
                )
            )

    # stop reason
    stop_reason = candidate_stop_reason(candidate.finish_reason)

    return ChatCompletionChoice(
        message=ChatMessageAssistant(
            content=content,
            tool_calls=tool_calls if len(tool_calls) > 0 else None,
            source="generate",
        ),
        stop_reason=stop_reason,
    )


def completion_choices_from_candidates(
    candidates: list[Candidate],
) -> list[ChatCompletionChoice]:
    candidates = copy(candidates)
    candidates.sort(key=lambda c: c.index)
    return [completion_choice_from_candidate(candidate) for candidate in candidates]


# google deson't export FinishReason (it's in a sub-namespace with a beta
# designation that seems destined to change, so we vendor the enum here)
class FinishReason:
    FINISH_REASON_UNSPECIFIED = 0
    STOP = 1
    MAX_TOKENS = 2
    SAFETY = 3
    RECITATION = 4
    OTHER = 5


def candidate_stop_reason(finish_reason: FinishReason) -> StopReason:
    match finish_reason:
        case FinishReason.STOP:
            return "stop"
        case FinishReason.MAX_TOKENS:
            return "length"
        case FinishReason.SAFETY | FinishReason.RECITATION:
            return "content_filter"
        case _:
            return "unknown"


def gapi_should_retry(ex: BaseException) -> bool:
    if isinstance(ex, Exception):
        return if_transient_error(ex)
    else:
        return False



import asyncio
import functools
import os
from dataclasses import dataclass
from queue import Empty, Queue
from threading import Thread
from typing import Any, Literal, Protocol, cast

import numpy as np
import torch
from torch import Tensor
from transformers import (  # type: ignore
    AutoModelForCausalLM,
    AutoTokenizer,
    PreTrainedTokenizerBase,
    set_seed,
)
from typing_extensions import override

from inspect_ai._util.constants import DEFAULT_MAX_TOKENS

from .._model import (
    ChatCompletionChoice,
    ChatMessage,
    ChatMessageAssistant,
    GenerateConfig,
    Logprob,
    Logprobs,
    ModelAPI,
    ModelOutput,
    ModelUsage,
    TopLogprob,
    simple_input_messages,
)
from .._tool import ToolChoice, ToolInfo
from .._util import chat_api_input


class HuggingFaceAPI(ModelAPI):
    def __init__(
        self,
        model_name: str,
        base_url: str | None = None,
        config: GenerateConfig = GenerateConfig(),
        **model_args: Any,
    ):
        super().__init__(model_name=model_name, base_url=base_url, config=config)

        # set random seeds
        if config.seed is not None:
            set_random_seeds(config.seed)

        # collect known model_args (then delete them so we can pass the rest on)
        def collect_model_arg(name: str) -> Any | None:
            nonlocal model_args
            value = model_args.get(name, None)
            if value:
                model_args.pop(name)
            return value

        device = collect_model_arg("device")
        tokenizer = collect_model_arg("tokenizer")
        model_path = collect_model_arg("model_path")
        tokenizer_path = collect_model_arg("tokenizer_path")
        self.batch_size = collect_model_arg("batch_size")
        self.chat_template = collect_model_arg("chat_template")

        # device
        if device:
            self.device = device
        elif torch.backends.mps.is_available():
            self.device = "mps"
        elif torch.cuda.is_available():
            self.device = "cuda:0"
        else:
            self.device = "cpu"

        # model
        if model_path:
            self.model = AutoModelForCausalLM.from_pretrained(
                model_path, device_map=self.device, **model_args
            )
        else:
            self.model = AutoModelForCausalLM.from_pretrained(
                model_name, device_map=self.device, **model_args
            )

        # tokenizer
        if tokenizer:
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)
        elif model_path:
            if tokenizer_path:
                self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)
            else:
                self.tokenizer = AutoTokenizer.from_pretrained(model_path)
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        # LLMs generally don't have a pad token and we need one for batching
        self.tokenizer.pad_token = self.tokenizer.eos_token
        self.tokenizer.padding_side = "left"

    async def generate(
        self,
        input: list[ChatMessage],
        tools: list[ToolInfo],
        tool_choice: ToolChoice,
        config: GenerateConfig,
    ) -> ModelOutput:
        # create chat
        chat = self.hf_chat(input)

        # prepare tokenizer
        tokenizer = functools.partial(self.tokenizer, return_tensors="pt", padding=True)

        # prepare generator
        kwargs: dict[str, Any] = dict(do_sample=True)
        if config.max_tokens is not None:
            kwargs["max_new_tokens"] = config.max_tokens
        if config.temperature is not None:
            kwargs["temperature"] = config.temperature
        if config.top_p is not None:
            kwargs["top_p"] = config.top_p
        if config.top_k is not None:
            kwargs["top_k"] = config.top_k
        if config.logprobs is not None:
            kwargs["output_logits"] = True
        if "return_dict_in_generate" in kwargs:
            assert kwargs["return_dict_in_generate"]
        kwargs["return_dict_in_generate"] = True
        generator = functools.partial(self.model.generate, **kwargs)

        # prepare decoder
        decoder = functools.partial(
            self.tokenizer.batch_decode,
            skip_special_tokens=True,
            clean_up_tokenization_spaces=False,
        )

        # generate (uses a queue to batch so we await)
        response = await batched_generate(
            GenerateInput(
                input=chat,
                device=self.device,
                tokenizer=tokenizer,
                generator=generator,
                decoder=decoder,
                batch_size=config.max_connections or self.max_connections(),
            )
        )

        # gather logprobs
        final_logprobs = None
        if config.logprobs is not None:
            final_logprobs = extract_logprobs(
                response=response,
                top_logprobs=config.top_logprobs,
                tokenizer=self.tokenizer,
            )

        # construct choice
        choice = ChatCompletionChoice(
            message=ChatMessageAssistant(
                content=response.output,
                source="generate",
            ),
            logprobs=Logprobs(content=final_logprobs)
            if final_logprobs is not None
            else None,
        )

        # return output
        return ModelOutput(
            model=self.model_name,
            choices=[choice],
            usage=ModelUsage(
                input_tokens=response.input_tokens,
                output_tokens=response.output_tokens,
                total_tokens=response.total_tokens,
            ),
        )

    @override
    def max_tokens(self) -> int | None:
        """Default is 16, bump it up to a value suitable for evals."""
        return DEFAULT_MAX_TOKENS

    @override
    def max_connections(self) -> int:
        """Effectively the batch size."""
        return 32

    def hf_chat(self, messages: list[ChatMessage]) -> str:
        # handle system message and consecutive user messages
        messages = simple_input_messages(messages)
        # convert to hf format
        hf_messages = chat_api_input(messages)
        # apply chat template
        chat = self.tokenizer.apply_chat_template(
            hf_messages,
            add_generation_prompt=True,
            tokenize=False,
            chat_template=self.chat_template,
        )
        # return
        return cast(str, chat)


def set_random_seeds(seed: int | None = None) -> None:
    if seed is None:
        seed = np.random.default_rng().integers(2**32 - 1)
    # python hash seed
    os.environ["PYTHONHASHSEED"] = str(seed)
    # transformers seed
    set_seed(seed)


class Tokenizer(Protocol):
    def __call__(
        self, input: list[str]
    ) -> dict[Literal["input_ids", "attention_mask"], Tensor]: ...


class Generator(Protocol):
    def __call__(self, input_ids: Tensor, attention_mask: Tensor) -> Tensor: ...


class Decoder(Protocol):
    def __call__(self, sequences: Tensor) -> list[str]: ...


@dataclass
class GenerateInput:
    input: str
    device: str
    tokenizer: Tokenizer
    generator: Generator
    decoder: Decoder
    batch_size: int


@dataclass
class GenerateOutput:
    output: str
    input_tokens: int
    output_tokens: int
    total_tokens: int
    logprobs: torch.Tensor | None


@dataclass
class _QueueItem:
    input: GenerateInput
    future: asyncio.Future[GenerateOutput]
    loop: asyncio.AbstractEventLoop


batch_thread: Thread | None = None

batch_queue: "Queue[_QueueItem]" = Queue()


async def batched_generate(input: GenerateInput) -> GenerateOutput:
    # start the background thread if necessary
    global batch_thread
    if batch_thread is None:
        batch_thread = Thread(target=process_batches, daemon=True)
        batch_thread.start()

    # enque the job
    loop = asyncio.get_event_loop()
    future: asyncio.Future[GenerateOutput] = loop.create_future()
    batch_queue.put(_QueueItem(input=input, future=future, loop=loop))

    # await the job
    await future

    # return it
    return future.result()


def process_batches() -> None:
    while True:
        # drain the queue (wait until no new messages have shown up for 2 secones)
        inputs: list[tuple[GenerateInput, asyncio.Future[GenerateOutput]]] = []
        while True:
            try:
                input = batch_queue.get(timeout=2)
                loop = input.loop
                inputs.append((input.input, input.future))
                if len(inputs) == input.input.batch_size:
                    # max batch size reached
                    break
            except Empty:
                # we have exhausted the queue
                break

        # see if we have any work to do
        if len(inputs) == 0:
            continue

        try:
            # capture the generator and decoder functions
            first_input = inputs[0][0]
            device = first_input.device
            tokenizer = first_input.tokenizer
            generator = first_input.generator
            decoder = first_input.decoder

            # tokenize and move to device
            tokenized_inputs = tokenizer([item[0].input for item in inputs])
            input_ids = tokenized_inputs["input_ids"]
            attention_mask = tokenized_inputs["attention_mask"]
            input_ids = input_ids.to(device)
            attention_mask = attention_mask.to(device)

            # generate
            with torch.inference_mode():
                generation_outputs = generator(
                    input_ids=input_ids, attention_mask=attention_mask
                )
                generate_ids = generation_outputs.sequences
                logits = generation_outputs.logits

            # get logprobs from logits
            logprobs = None
            if logits is not None:
                logits = torch.stack(logits).transpose(0, 1)
                logprobs = torch.nn.functional.log_softmax(logits, dim=-1)

            # decode
            generated_tokens = generate_ids[:, input_ids.size(dim=1) :]
            if logprobs is not None:
                assert logprobs.shape[1] == generated_tokens.shape[1]
            outputs = decoder(sequences=generated_tokens)

            # call back futures
            for i, output in enumerate(outputs):
                future = inputs[i][1]
                input_tokens = input_ids.size(dim=1)
                output_tokens = generate_ids.size(dim=1) - input_ids.size(dim=1)

                # asyncio futures are not thread safe, so we need to pass the event loop
                # down to this point, so we can mark the future as done in a thread safe manner.
                # see: https://docs.python.org/3/library/asyncio-dev.html#concurrency-and-multithreading
                loop.call_soon_threadsafe(
                    future.set_result,
                    GenerateOutput(
                        output=output,
                        input_tokens=input_tokens,
                        output_tokens=output_tokens,
                        total_tokens=input_tokens + output_tokens,
                        logprobs=logprobs[i] if logprobs is not None else None,
                    ),
                )

        except Exception as ex:
            for inp in inputs:
                future = inp[1]
                loop.call_soon_threadsafe(future.set_exception, ex)


def extract_logprobs(
    response: GenerateOutput,
    top_logprobs: int | None,
    tokenizer: PreTrainedTokenizerBase,
) -> list[Logprob]:
    assert response.logprobs is not None
    k = top_logprobs or 1
    topk_values, topk_inds = response.logprobs.topk(k=k, dim=-1)
    final_logprobs = []
    for toks, vals in zip(topk_inds, topk_values):
        top_logprobs = []
        for tok, val in zip(toks, vals):
            # TODO: you get byte artifacts converting single ids to tokens like this...
            # but `tokenizer.decode` strips spaces. There must be a better way to do this.
            token_str = tokenizer.convert_ids_to_tokens(tok.item())
            top_logprobs.append(
                TopLogprob(
                    token=token_str,
                    logprob=val,
                    bytes=list(map(ord, token_str)),
                )
            )
        final_logprobs.append(
            Logprob(
                token=top_logprobs[0].token,
                logprob=top_logprobs[0].logprob,
                bytes=top_logprobs[0].bytes,
                top_logprobs=top_logprobs,
            )
        )
    return final_logprobs



from typing import Any

from utils import skip_if_no_openai

from inspect_ai import Task, eval, score
from inspect_ai._util.constants import PKG_NAME
from inspect_ai._util.registry import registry_info
from inspect_ai.dataset import Sample
from inspect_ai.scorer import Metric, Score, accuracy, includes, match, metric
from inspect_ai.scorer._metric import MetricType, metric_create

# declare some metrics using the various forms supported (function,
# function returning Metric, class deriving from Metric) as well
# as using implicit and explicit names


@metric
def accuracy1(correct: str = "C") -> Metric:
    def metric(scores: list[Score]) -> int | float:
        return 1

    return metric


@metric(name="accuracy2")
def acc_fn(correct: str = "C") -> Metric:
    def metric(scores: list[Score]) -> int | float:
        return 1

    return metric


@metric
class Accuracy3(Metric):
    def __init__(self, correct: str = "C") -> None:
        self.correct = correct

    def __call__(self, scores: list[Score]) -> int | float:
        return 1


@metric(name="accuracy4")
class AccuracyNamedCls(Metric):
    def __init__(self, correct: str = "C") -> None:
        self.correct = correct

    def __call__(self, scores: list[Score]) -> int | float:
        return 1


def test_metric_registry() -> None:
    registry_assert(accuracy1, "accuracy1")
    registry_assert(acc_fn, "accuracy2")
    registry_assert(Accuracy3, "Accuracy3")
    registry_assert(AccuracyNamedCls, "accuracy4")


def test_metric_call() -> None:
    registry_assert(accuracy1(), "accuracy1")
    registry_assert(acc_fn(), "accuracy2")
    registry_assert(Accuracy3(), "Accuracy3")
    registry_assert(AccuracyNamedCls(), "accuracy4")


def test_metric_create() -> None:
    metric_create_assert("accuracy1", correct="C")
    metric_create_assert("accuracy1", correct="C")
    metric_create_assert("Accuracy3", correct="C")
    metric_create_assert("accuracy4", correct="C")


def test_inspect_metrics() -> None:
    registry_assert(accuracy, f"{PKG_NAME}/accuracy")
    registry_assert(accuracy(), f"{PKG_NAME}/accuracy")


@skip_if_no_openai
def test_extra_metrics() -> None:
    # check that we get the extra metrics and de-duping works
    def check_log(log):
        assert log.results and (
            list(log.results.metrics.keys())
            == [
                "accuracy",
                "bootstrap_std",
                "accuracy1",
                "Accuracy3",
            ]
        )

    task = Task(
        dataset=[Sample(input="What is 1 + 1?", target=["2", "2.0", "Two"])],
        scorer=match(),
        metrics=[accuracy(), accuracy1(), Accuracy3()],
    )

    # normal eval
    log = eval(task)[0]
    check_log(log)

    # eval log w/ different scorer (that still uses accuracy)
    log = score(log, scorer=includes())
    check_log(log)


def registry_assert(metric: Metric | MetricType, name: str) -> None:
    info = registry_info(metric)
    assert info.name == name


def metric_create_assert(name: str, **kwargs: Any) -> None:
    metric = metric_create(name, **kwargs)
    assert metric([]) == 1



from pathlib import Path

from platformdirs import user_runtime_dir

from inspect_ai._util.constants import PKG_NAME


def inspect_runtime_dir(subdir: str | None) -> Path:
    runtime_dir = Path(user_runtime_dir(PKG_NAME))
    if subdir:
        runtime_dir = runtime_dir / subdir
    runtime_dir.mkdir(parents=True, exist_ok=True)
    return runtime_dir



# ruff: noqa: F403 F405

from ._dataset import (
    Dataset,
    FieldSpec,
    MemoryDataset,
    RecordToSample,
    Sample,
)
from ._sources.csv import csv_dataset
from ._sources.example import example_dataset
from ._sources.file import file_dataset
from ._sources.hf import hf_dataset
from ._sources.json import json_dataset

__all__ = [
    "Dataset",
    "Sample",
    "FieldSpec",
    "RecordToSample",
    "MemoryDataset",
    "file_dataset",
    "csv_dataset",
    "hf_dataset",
    "json_dataset",
    "example_dataset",
]



import asyncio

from inspect_ai._util.images import image_as_data_uri
from inspect_ai.dataset import Sample
from inspect_ai.model import ChatMessage, ChatMessageUser, Content, ContentImage


async def samples_with_base64_images(samples: list[Sample]) -> list[Sample]:
    return await asyncio.gather(
        *[sample_with_base64_images(sample) for sample in samples]
    )


async def sample_with_base64_images(sample: Sample) -> Sample:
    if isinstance(sample.input, list):
        return Sample(
            input=await messages_with_base64_images(sample.input),
            target=sample.target,
            id=sample.id,
            metadata=sample.metadata,
        )
    else:
        return sample


async def messages_with_base64_images(messages: list[ChatMessage]) -> list[ChatMessage]:
    return await asyncio.gather(
        *[message_with_base64_image(message) for message in messages]
    )


async def message_with_base64_image(message: ChatMessage) -> ChatMessage:
    if isinstance(message, ChatMessageUser) and not isinstance(message.content, str):
        return ChatMessageUser(
            content=[
                await chat_content_with_base64_image(content)
                for content in message.content
            ],
            source=message.source,
        )
    else:
        return message


async def chat_content_with_base64_image(content: Content) -> Content:
    if isinstance(content, ContentImage):
        if isinstance(content.image, str):
            return ContentImage(image=await image_as_data_uri(content.image))
        else:
            return ContentImage(
                image=await image_as_data_uri(content.image.url),
                detail=content.image.detail,
            )
    else:
        return content



from typing import Any, cast

from inspect_langchain import langchain_solver
from langchain import hub
from langchain.agents import (
    AgentExecutor,
    BaseMultiActionAgent,
    create_openai_tools_agent,
    load_tools,
)
from langchain.tools.tavily_search import TavilySearchResults
from langchain.utilities.tavily_search import TavilySearchAPIWrapper
from langchain_core.language_models import BaseChatModel

from inspect_ai import Task, task
from inspect_ai.dataset import json_dataset
from inspect_ai.scorer import model_graded_fact
from inspect_ai.solver import Solver, solver


@solver
def wikipedia_search(
    max_iterations: int | None = 15, max_execution_time: float | None = None
) -> Solver:
    # standard prompt for functions agent
    prompt = hub.pull("hwchase17/openai-tools-agent")

    # tavily and wikipedia tools
    tavily_api = TavilySearchAPIWrapper()  # type: ignore
    tools = [TavilySearchResults(api_wrapper=tavily_api)] + load_tools(["wikipedia"])

    # agent function
    async def agent(llm: BaseChatModel, input: dict[str, Any]):
        # create agent -- cast needed due to:
        # https://github.com/langchain-ai/langchain/issues/13075
        tools_agent = create_openai_tools_agent(llm, tools, prompt)
        agent_executor = AgentExecutor.from_agent_and_tools(
            agent=cast(BaseMultiActionAgent, tools_agent),
            tools=tools,
            name="wikipedia_search",
            max_iterations=max_iterations,
            max_execution_time=max_execution_time,
        )

        # execute the agent and return output
        result = await agent_executor.ainvoke(input)
        return result["output"]

    # return agent function as inspect solver
    return langchain_solver(agent)


@task
def wikipedia() -> Task:
    return Task(
        dataset=json_dataset("wikipedia.jsonl"),
        plan=wikipedia_search(),
        scorer=model_graded_fact(),
    )



import asyncio

import click
from typing_extensions import Unpack

from inspect_ai._display import display
from inspect_ai._display.logger import init_logger
from inspect_ai._eval.loader import load_tasks
from inspect_ai._util.constants import SCORED_SUFFIX
from inspect_ai._util.dotenv import init_dotenv
from inspect_ai.log._file import JSONRecorder
from inspect_ai.model import get_model
from inspect_ai.model._model import init_async_context_model
from inspect_ai.util._context import init_async_context

from .common import CommonOptions, common_options, resolve_common_options


@click.command("score")
@click.argument("task", type=str)
@click.argument("log-file", type=str, required=False)
@click.option(
    "--no-overwrite",
    type=bool,
    is_flag=True,
    help="Do not overwrite unscored log_files with the scored version (instead write a new file w/ '-scored' appended)",
)
@common_options
def score_command(
    task: str,
    log_file: str | None,
    no_overwrite: bool | None,
    **kwargs: Unpack[CommonOptions],
) -> None:
    """Score a previous evaluation run."""
    # read common options
    (log_dir, log_level) = resolve_common_options(kwargs)

    # score
    asyncio.run(
        score(task, log_dir, log_file, False if no_overwrite else True, log_level)
    )


async def score(
    task: str,
    log_dir: str,
    log_file: str | None,
    overwrite: bool,
    log_level: str | None,
) -> None:
    init_dotenv()
    init_logger(log_level)

    # read the eval log
    recorder = JSONRecorder(log_dir)
    log_file = log_file if log_file else recorder.latest_log_file_path()
    eval_log = recorder.read_log(log_file)

    # check that there are samples therein
    if eval_log.samples is None or len(eval_log.samples) == 0:
        raise ValueError(f"{log_file} does not include samples to score")

    # get the model then initialize the async context
    model = get_model(
        model=eval_log.eval.model,
        config=eval_log.plan.config,
        **eval_log.eval.model_args,
    )

    # initialize async contexts
    init_async_context()
    init_async_context_model(model)

    # instantiate the task so we can get its scorer and metrics
    score_task = load_tasks([task], model)[0]

    # re-score the task
    eval_log = await score_task.score(eval_log)

    # re-write the log (w/ a -score suffix if requested)
    scored = f"{SCORED_SUFFIX}.json"
    if not overwrite and not log_file.endswith(scored):
        log_file = log_file.removesuffix(".json") + scored
    recorder.write_log(log_file, eval_log)

    # print results
    display().print(f"\n{eval_log.eval.task}")
    if eval_log.results:
        for name, metric in eval_log.results.metrics.items():
            display().print(f"{name}: {metric.value}")
    display().print(f"log: {log_file}\n")



from utils import skip_if_no_openai

from inspect_ai import Task, eval
from inspect_ai.dataset import Sample
from inspect_ai.model import ChatMessageUser, ModelOutput, get_model
from inspect_ai.scorer import match
from inspect_ai.solver import (
    Generate,
    Plan,
    TaskState,
    chain_of_thought,
    generate,
    solver,
)


@skip_if_no_openai
def test_solvers_termination():
    @solver
    def user_input(input: str):
        async def solve(state: TaskState, generate: Generate):
            state.messages.append(ChatMessageUser(content=input))
            return state

        return solve

    @solver
    def complete_task():
        async def solve(state: TaskState, generate: Generate):
            state.completed = True
            return state

        return solve

    @solver
    def finish():
        async def solve(state: TaskState, generate: Generate):
            state.output = ModelOutput.from_content(
                model="openai/gpt-4", content="finished"
            )
            return state

        return solve

    model = get_model("openai/gpt-4")
    task = Task(
        dataset=[Sample(input="What is 1 + 1?", target=["2", "2.0", "Two"])],
        plan=Plan(
            steps=[
                chain_of_thought(),
                generate(),
                user_input("How about multiplying the numbers?"),
                generate(),
                complete_task(),
                user_input("How about subtracting the numbers?"),
                generate(),
            ],
            finish=finish(),
        ),
        scorer=match(),
    )

    log = eval(task, model=model)[0]
    assert len(log.samples[0].messages) == 4
    assert log.samples[0].output.completion == "finished"

    log = eval(task, model=model, max_messages=2)[0]
    assert len(log.samples[0].messages) == 2
    assert log.samples[0].output.completion == "finished"



from docstring_parser import Docstring, parse


def parse_docstring(
    docstring: str | None,
) -> Docstring:
    if docstring is None:
        return Docstring()
    parsed_docstring = parse(docstring)
    if parsed_docstring.short_description is None:
        raise ValueError("Docstring must have a short description")
    return parsed_docstring



# ruff: noqa: F401 F403 F405

from ._model import (
    ChatCompletionChoice,
    ChatMessage,
    ChatMessageAssistant,
    ChatMessageSystem,
    ChatMessageTool,
    ChatMessageUser,
    Content,
    ContentImage,
    ContentText,
    GenerateConfig,
    GenerateConfigArgs,
    Logprob,
    Logprobs,
    Model,
    ModelAPI,
    ModelName,
    ModelOutput,
    ModelUsage,
    StopReason,
    TopLogprob,
    get_model,
)
from ._providers.providers import *
from ._registry import modelapi
from ._tool import ToolCall, ToolChoice, ToolFunction, ToolInfo, ToolParam

__all__ = [
    "GenerateConfig",
    "GenerateConfigArgs",
    "ContentText",
    "ContentImage",
    "Content",
    "ChatMessage",
    "ChatMessageSystem",
    "ChatMessageUser",
    "ChatMessageAssistant",
    "ChatMessageTool",
    "ChatCompletionChoice",
    "ModelOutput",
    "Logprobs",
    "Logprob",
    "TopLogprob",
    "Model",
    "ModelAPI",
    "ModelName",
    "ModelUsage",
    "StopReason",
    "ToolCall",
    "ToolChoice",
    "ToolFunction",
    "ToolInfo",
    "ToolParam",
    "ToolType",
    "get_model",
    "modelapi",
]



import logging
from typing import Callable

from httpx import ConnectError, ConnectTimeout, HTTPStatusError, ReadTimeout
from tenacity import RetryCallState

from inspect_ai._util.constants import HTTP

logger = logging.getLogger(__name__)


def httpx_should_retry(ex: BaseException) -> bool:
    """Check whether an exception raised from httpx should be retried.

    Implements the strategy described here: https://cloud.google.com/storage/docs/retry-strategy

    Args:
      ex (BaseException): Exception to examine for retry behavior

    Returns:
      True if a retry should occur
    """
    # httpx status exception
    if isinstance(ex, HTTPStatusError):
        # request timeout
        if ex.response.status_code == 408:
            return True
        # lock timeout
        elif ex.response.status_code == 409:
            return True
        # rate limit
        elif ex.response.status_code == 429:
            return True
        # internal errors
        elif ex.response.status_code >= 500:
            return True
        else:
            return False

    # connection error
    elif is_httpx_connection_error(ex):
        return True

    # don't retry
    else:
        return False


def log_rate_limit_retry(context: str, retry_state: RetryCallState) -> None:
    logger.log(
        HTTP,
        f"{context} rate limit retry {retry_state.attempt_number} after waiting for {retry_state.idle_for}",
    )


def log_retry_attempt(context: str) -> Callable[[RetryCallState], None]:
    def log_attempt(retry_state: RetryCallState) -> None:
        logger.log(
            HTTP,
            f"{context} connection retry {retry_state.attempt_number} after waiting for {retry_state.idle_for}",
        )

    return log_attempt


def is_httpx_connection_error(ex: BaseException) -> bool:
    if (
        isinstance(ex, ConnectTimeout)
        or isinstance(ex, ConnectError)
        or isinstance(ex, ConnectionError)
        or isinstance(ex, ReadTimeout)
    ):
        return True
    else:
        return False



from typing import (
    Any,
    Callable,
    Protocol,
    TypeVar,
    cast,
    overload,
    runtime_checkable,
)

from typing_extensions import Unpack

from inspect_ai._util.registry import (
    RegistryInfo,
    registry_add,
    registry_create,
    registry_name,
    registry_tag,
)
from inspect_ai.model import (
    ChatMessage,
    ChatMessageUser,
    GenerateConfigArgs,
    ModelName,
    ModelOutput,
    ToolChoice,
)

from ._tool.tool import Tool


class TaskState:
    def __init__(
        self,
        model: ModelName,
        sample_id: int | str,
        epoch: int,
        input: str | list[ChatMessage],
        choices: list[str] | None,
        messages: list[ChatMessage],
        tools: list[Tool] = [],
        tool_choice: ToolChoice | None = None,
        output: ModelOutput | None = None,
        completed: bool = False,
        metadata: dict[str, Any] = {},
    ) -> None:
        self._model = model

        self.sample_id = sample_id
        """Unique id for sample."""

        self.epoch = epoch
        """Epoch number for sample."""

        self._input = input

        self.choices = choices
        """Sample choices."""

        self.messages = messages
        """Chat conversation history for sample."""

        self.tools = tools
        """Tools available to the model."""

        self.tool_choice = tool_choice
        """Tool choice directive."""

        self.output = output if output else ModelOutput(model=str(model), choices=[])
        """Model output."""

        self.completed = completed
        """Flag to indicate that the solver loop should terminate."""

        self.metadata = metadata
        """Additional task state metadata."""

    @property
    def model(self) -> ModelName:
        """Name of model being evaluated."""
        return self._model

    @property
    def input(self) -> str | list[ChatMessage]:
        """Sample input."""
        return self._input

    @property
    def input_text(self) -> str:
        """Sample input as text."""
        if isinstance(self._input, str):
            return self._input
        else:
            return next(
                (message.text for message in self.messages if message.role == "user"),
                "",
            )

    @property
    def user_prompt(self) -> ChatMessageUser:
        """User prompt for this state.

        Tasks are very general and can have may types of inputs.
        However, in many cases solvers assume they can interact with
        the state as a "chat" in a predictable fashion (e.g. prompt
        engineering solvers). This propery enables easy read and
        write access to the user chat prompt. Raises an
        exception if there is no user prompt

        Returns:
           First user `ChatMessage` if the current state has one, else `None`
        """
        prompt = next(
            (m for m in self.messages if isinstance(m, ChatMessageUser)), None
        )
        if prompt:
            return prompt
        else:
            raise ValueError("User prompt requested from TaskState but none available")


@runtime_checkable
class Generate(Protocol):
    """Generate using the model and add the assistant message to the task state.

    Args:
       state (TaskState): Beginning task state.
       **kwargs: Optional generation config arguments.

    Returns:
       Updated TaskState.
    """

    async def __call__(
        self, state: TaskState, **kwargs: Unpack[GenerateConfigArgs]
    ) -> TaskState: ...


@runtime_checkable
class Solver(Protocol):
    r"""Contribute to solving an evaluation task.

    Contribute to the solution of a task by transforming a TaskState
    (e.g. prompt enhancement, eliciation, etc.). Solvers return a
    TaskState (which could simply be a modified version of the one
    they were passed) and optionally may call the generate() function
    to generate output (and a new TaskState with that output).


    Args:
        state (TaskState): States for tasks being evaluated.
        generate (Generate): Function for generating outputs.

    Returns:
        Updated TaskState.
    """

    async def __call__(
        self,
        state: TaskState,
        generate: Generate,
    ) -> TaskState: ...


SolverType = TypeVar("SolverType", Callable[..., Solver], type[Solver])
r"""Solver type.

Valid solver types include:
 - Functions that return a Solver
 - Classes derivied from Solver
"""


def solver_register(solver: SolverType, name: str = "") -> SolverType:
    r"""Register a function or class as a solver.

    Args:
        solver (SolverType):
            Function that returns a Solver or class derived Solver.
        name (str): Name of solver (Optional, defaults to object name)

    Returns:
        Solver with registry attributes.
    """
    solver_name = name if name else getattr(solver, "__name__")
    registry_add(solver, RegistryInfo(type="solver", name=solver_name))
    return solver


def solver_create(name: str, **kwargs: Any) -> Solver:
    r"""Create a Solver based on its registered name.

    Args:
        name (str): Name of solver (Optional, defaults to object name)
        **kwargs (dict): Optional creation arguments for the solver

    Returns:
        Solver with registry info attribute
    """
    return cast(Solver, registry_create("solver", name, **kwargs))


@overload
def solver(name: str) -> Callable[..., SolverType]: ...


@overload
# type: ignore
def solver(name: Callable[..., Solver]) -> Callable[..., Solver]: ...


@overload
def solver(name: type[Solver]) -> type[Solver]: ...


def solver(name: str | SolverType) -> Callable[..., SolverType] | SolverType:
    r"""Decorator for registering solvers.

    Args:
        name: (str | SolverType):
            Optional name for solver. If the decorator has no name
            argument then the name of the underlying SolverType
            object will be used to automatically assign a name.

    Returns:
        Solver with registry attributes.

    Exmaples:
        @solver
        def prompt_cot(state: TaskState, generate: Generate) -> None:
            ...

        @solver(name = "prompt_cot")
        def cot(state: TaskState, generate: Generate) -> None:
            ...

        @solver
        def prompt_cot(template: str) -> Solver:
            def solve(state: TaskState, generate: Generate) -> None:
                ...
            return solve
    """

    # create_solver_wrapper:
    #  (a) Add the SolverType to the registry using the appropriately
    #      package-namespaced name
    #  (b) Ensure that instances of Solver created by SolverType also
    #      carry registry info.
    def create_solver_wrapper(
        solver_type: SolverType, name: str | None = None
    ) -> SolverType:
        solver_name = registry_name(
            solver_type, name if name else getattr(solver_type, "__name__")
        )

        def solver_wrapper(*args: Any, **kwargs: dict[str, Any]) -> Solver:
            solver = solver_type(*args, **kwargs)

            registry_tag(
                solver_type,
                solver,
                RegistryInfo(type="solver", name=solver_name),
                *args,
                **kwargs,
            )

            return solver

        return solver_register(cast(SolverType, solver_wrapper), solver_name)

    # for decorators with an explicit name, one more wrapper for the name
    if isinstance(name, str):

        def wrapper(solver_type: SolverType) -> SolverType:
            return create_solver_wrapper(solver_type, name)

        return wrapper

    # create a solver wrapper for the passsed solver_type
    else:
        solver_type = name
        return create_solver_wrapper(solver_type)


@solver
def generate() -> Solver:
    r"""Generate output from the model and append it to task message history.

    generate() is the default plan/solver if none is specified for a given task.
    """

    # call generate on the tasks
    async def solve(state: TaskState, generate: Generate) -> TaskState:
        return await generate(state)

    # return solve
    return solve



from inspect_ai import Task, task
from inspect_ai.dataset import Sample
from inspect_ai.scorer import includes, match
from inspect_ai.solver import generate, system_message, tool, use_tools
from inspect_ai.util import subprocess


@tool(prompt="""
    If you are given a math problem of any kind,
    please use the add tool to compute the result.
    """
)
def add():
    async def execute(x: int, y: int):
        """
        Tool for adding two numbers.

        Args:
            x (int): First number to add.
            y (int): Second number to add.

        Returns:
            The sum of the two numbers.
        """
        return x + y

    return execute

@task
def addition_problem():
    return Task(
        dataset=[Sample(
            input="What is 1 + 1?",
            target=["2", "2.0"]
        )],
        plan=[use_tools(add()), generate()],
        scorer=match(numeric=True),
    )

@tool(
    prompt="""
    If you are asked to list the files in a directory you
    should call the list_files function to list the files.
    """
)
def list_files():
    async def execute(dir: str):
        """List the files in a directory.

        Args:
            dir (str): Directory

        Returns:
            File listing of the directory
        """
        result = await subprocess(["ls", dir])
        if result.success:
            return result.stdout
        else:
            return f"Error: {result.stderr}"

    return execute

SYSTEM_MESSAGE = """
Please answer exactly Yes or No with no additional words.
"""

@task
def bash():

    dataset = [Sample(
        input=(
            "Please list the files in the /usr/bin directory. "
            + "Is there a file named 'python3' in the directory?"
        ),
        target=["Yes"],
    )]

    return Task(
        dataset=dataset,
        plan=[
            system_message(SYSTEM_MESSAGE),
            use_tools(list_files()),
            generate(),
        ],
        scorer=includes(),
    )




from .concurrency import init_concurrency
from .logger import init_logger_records
from .subprocess import init_subprocess


def init_async_context(max_subprocesses: int | None = None) -> None:
    init_concurrency()
    init_subprocess(max_subprocesses)
    init_logger_records()



from inspect_ai import Task, task
from inspect_ai.dataset import example_dataset
from inspect_ai.scorer import model_graded_fact
from inspect_ai.solver import generate, system_message

SYSTEM_MESSAGE = """
You are a computer security expert and you are tasked with providing concise response to the following questions. Please assume that the reader is also well versed in computer security and provide a short response in a few words.
"""


@task
def security_guide():
    return Task(
        dataset=example_dataset("security_guide"),
        plan=[system_message(SYSTEM_MESSAGE), generate()],
        scorer=model_graded_fact(),
    )



from ._critique import self_critique
from ._multiple_choice import multiple_choice
from ._plan import Plan, plan
from ._prompt import (
    chain_of_thought,
    prompt_template,
    system_message,
)
from ._solver import Generate, Solver, TaskState, generate, solver
from ._tool.tool import Tool, tool
from ._tool.use_tools import use_tools
from ._tool.web_search import web_search

__all__ = [
    "generate",
    "prompt_template",
    "chain_of_thought",
    "multiple_choice",
    "system_message",
    "self_critique",
    "tool",
    "use_tools",
    "web_search",
    "plan",
    "Plan",
    "Solver",
    "solver",
    "TaskState",
    "Tool",
    "Generate",
]



import functools
from typing import Any, Callable, Tuple, cast

import click
from typing_extensions import TypedDict

from inspect_ai._util.constants import DEFAULT_LOG_LEVEL


class CommonOptions(TypedDict):
    log_level: str
    log_dir: str
    debug: bool
    debug_port: int


def common_options(func: Callable[..., Any]) -> Callable[..., click.Context]:
    @click.option(
        "--log-level",
        type=click.Choice(
            ["debug", "http", "info", "warning", "error", "critical"],
            case_sensitive=False,
        ),
        default=DEFAULT_LOG_LEVEL,
        envvar="INSPECT_LOG_LEVEL",
        help=f"Set the log level (defaults to '{DEFAULT_LOG_LEVEL}')",
    )
    @click.option(
        "--log-dir",
        type=str,
        default="./logs",
        envvar="INSPECT_LOG_DIR",
        help="Directory for log files.",
    )
    @click.option(
        "--debug", is_flag=True, envvar="INSPECT_DEBUG", help="Wait to attach debugger"
    )
    @click.option(
        "--debug-port",
        default=5678,
        envvar="INSPECT_DEBUG_PORT",
        help="Port number for debugger",
    )
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> click.Context:
        return cast(click.Context, func(*args, **kwargs))

    return wrapper


def resolve_common_options(options: CommonOptions) -> Tuple[str, str]:
    # attach debugger if requested
    if options["debug"]:
        import debugpy  # type: ignore

        debugpy.listen(options["debug_port"])
        print("Waiting for debugger attach")
        debugpy.wait_for_client()
        print("Debugger attached")

    # return resolved options
    return (options["log_dir"], options["log_level"])



from pathlib import Path

from .._dataset import Dataset, FieldSpec, MemoryDataset, RecordToSample
from .csv import csv_dataset
from .json import json_dataset

EXAMPLES_PATH = Path(__file__).parent.parent / "_examples"


def example_dataset(
    name: str,
    sample_fields: FieldSpec | RecordToSample | None = None,
) -> Dataset:
    """Read a dataset from inspect_ai package examples.

    This is primarily used for sharing runnable example
    snippets that don't need to read an external dataset.

    Args:
      name (str): Example dataset name. One of 'security_guide', 'theory_of_mind',
        'popularity', or 'biology_qa'
      sample_fields (SampleFieldSpec | RecordToSample): Method of mapping underlying
        fields in the data source to `Sample` objects. Pass `None` if the data is already
        stored in `Sample` form (i.e. object with "input" and "target" fields); Pass a
        `SampleFieldSpec` to specify mapping fields by name; Pass a `RecordToSample` to
        handle mapping with a custom function.


    Returns:
      Dataset read from example file.
    """
    json_file = (EXAMPLES_PATH / f"{name}.jsonl").as_posix()
    csv_file = (EXAMPLES_PATH / f"{name}.csv").as_posix()
    if not Path(json_file).exists() and Path(csv_file).exists():
        raise ValueError(f"Sample dataset {name} not found.")

    if Path(json_file).exists():
        dataset = json_dataset(
            json_file=json_file,
            sample_fields=sample_fields,
        )
    else:
        dataset = csv_dataset(
            csv_file=csv_file,
            sample_fields=sample_fields,
        )

    return MemoryDataset(samples=list(dataset), name=name, location=f"example://{name}")



import pytest
from utils import (
    skip_if_no_anthropic,
    skip_if_no_mistral,
    skip_if_no_openai,
    skip_if_no_together,
)

from inspect_ai.model import GenerateConfig, ModelOutput, get_model


async def generate(model_name) -> ModelOutput:
    model = get_model(model_name)
    return await model.generate(input="Hello.")


async def generate_token_limit(model_name) -> ModelOutput:
    model = get_model(model_name)
    return await model.generate(
        input="Tell me a story.", config=GenerateConfig(max_tokens=2)
    )


async def check_stop_reason(model_name):
    response = await generate(model_name)
    assert response.choices[0].stop_reason == "stop"

    response = await generate_token_limit(model_name)
    assert response.choices[0].stop_reason == "length"


@pytest.mark.asyncio
@skip_if_no_openai
async def test_openai_stop_reason() -> None:
    await check_stop_reason("openai/gpt-3.5-turbo")


@pytest.mark.asyncio
@skip_if_no_anthropic
async def test_anthropic_stop_reason() -> None:
    await check_stop_reason("anthropic/claude-3-haiku-20240307")


@pytest.mark.asyncio
@skip_if_no_mistral
async def test_mistral_stop_reason() -> None:
    await check_stop_reason("mistral/mistral-medium-latest")


@pytest.mark.asyncio
@skip_if_no_together
async def test_together_stop_reason() -> None:
    await check_stop_reason("together/google/gemma-2b-it")



# mypy: disable-error-code="unused-ignore"

from pathlib import Path
from typing import Any

from inspect_ai._util.error import pip_dependency_error
from inspect_ai._util.version import verify_required_version

from .._dataset import (
    Dataset,
    FieldSpec,
    MemoryDataset,
    RecordToSample,
)
from .._util import record_to_sample_fn


def hf_dataset(
    path: str,
    name: str | None = None,
    data_dir: str | None = None,
    split: str | None = None,
    sample_fields: FieldSpec | RecordToSample | None = None,
    shuffle: bool = False,
    seed: int | None = None,
    limit: int | None = None,
    trust: bool = False,
    **kwargs: dict[str, Any],
) -> Dataset:
    """Datasets read using the Hugging Face `datasets` package.

    The `hf_dataset` function supports reading datasets using the Hugging Face
    `datasets` package, including remote datasets on Hugging Face Hub.

    Args:
        path (str): Path or name of the dataset. Depending on path, the dataset
          builder that is used comes from a generic dataset script (JSON, CSV,
          Parquet, text etc.) or from the dataset script (a python file) inside
          the dataset directory.
        name (str | None): Name of the dataset configuration.
        data_dir (str | None): data_dir of the dataset configuration
          to read data from.
        split (str | None): Which split of the data to load.
        sample_fields (SampleFieldSpec | RecordToSample): Method of mapping underlying
          fields in the data source to Sample objects. Pass `None` if the data is already
          stored in `Sample` form (i.e. has "input" and "target" columns.); Pass a
          `SampleFieldSpec` to specify mapping fields by name; Pass a `RecordToSample` to
          handle mapping with a custom function.
        shuffle (bool): Randomly shuffle the dataset order.
        seed: (int | None): Seed used for random shuffle.
        limit (int | None): Limit the number of records to read.
        trust (bool): Whether or not to allow for datasets defined on the Hub
          using a dataset script. This option should only be set to True for
          repositories you trust and in which you have read the code, as it
          will execute code present on the Hub on your local machine.
        **kwargs (dict[str, Any]): Additional arguments to pass through to the
          `load_dataset` function of the `datasets` package.

    Returns:
        Dataset read from Hugging Face
    """
    # ensure we have the datasets package (>= v2.16, which supports trust_remote_code)
    FEATURE = "Hugging Face Datasets"
    PACKAGE = "datasets"
    VERSION = "2.16.0"
    try:
        import datasets  # type: ignore
    except ImportError:
        raise pip_dependency_error(FEATURE, [PACKAGE])
    verify_required_version(FEATURE, PACKAGE, VERSION)

    # resolve data_to_sample function
    data_to_sample = record_to_sample_fn(sample_fields)

    # load the dataset as a list of dicts
    dataset = datasets.load_dataset(  # type: ignore
        path=path,
        name=name,
        data_dir=data_dir,
        split=split,
        trust_remote_code=trust,
        **kwargs,
    )

    # shuffle if requested
    if shuffle:
        dataset.shuffle(seed=seed)

    # limit if requested
    if limit:
        dataset = dataset.select(range(limit))

    # return the dataset
    return MemoryDataset(
        samples=[data_to_sample(data) for data in dataset.to_list()],
        name=Path(path).stem if Path(path).exists() else path,
        location=path,
    )



from logging import INFO, LogRecord

_logger_records: list[LogRecord] = []
_rate_limit_records: list[LogRecord] = []


def init_logger_records() -> None:
    _logger_records.clear()
    _rate_limit_records.clear()


def notify_logger_record(record: LogRecord, write: bool) -> None:
    if write:
        _logger_records.append(record)
    if record.levelno <= INFO and "429" in record.getMessage():
        _rate_limit_records.append(record)


def logger_http_rate_limit_count() -> int:
    return len(_rate_limit_records)


def collect_logger_records() -> list[LogRecord]:
    records = _logger_records.copy()
    _logger_records.clear()
    _rate_limit_records.clear()
    return records



from typing import Literal

from ._common import match_str, str_match_scorer
from ._metrics import accuracy, bootstrap_std
from ._scorer import Scorer, scorer


@scorer(metrics=[accuracy(), bootstrap_std()])
def match(
    location: Literal["begin", "end", "any", "exact"] = "end",
    *,
    ignore_case: bool = True,
    numeric: bool = False,
) -> Scorer:
    """Scorer which matches text or a number.

    Args:
       location (Literal["begin", "end", "any", "exact"]):
          Location to match at. "any" matches anywhere in the
          output; "exact" requires the output be exactly
          equal to the target (module whitespace, etc.)
       ignore_case (bool): Do case insenstive comparison.
       numeric (bool): Is this a numeric match? (in this
          case different punctuation removal rules are
          used and numbers are normalized before comparisoin).
    """

    def check(value: str, target: str) -> tuple[str, bool]:
        return match_str(
            value=value,
            target=target,
            location=location,
            ignore_case=ignore_case,
            numeric=numeric,
        )

    return str_match_scorer(check)


@scorer(metrics=[accuracy(), bootstrap_std()])
def includes(ignore_case: bool = True) -> Scorer:
    """Check whether the specified text is included in the model output.

    Args:
       ignore_case (bool): Use a case insensitive comparison.

    """

    def check(value: str, target: str) -> tuple[str, bool]:
        if ignore_case:
            idx = value.lower().rfind(target.lower())
        else:
            idx = value.rfind(target)
        return value, idx != -1

    return str_match_scorer(check)



from enum import Enum
from typing import Literal

from inspect_ai._util.pattern import (
    ANSWER_PATTERN_LETTER,
    ANSWER_PATTERN_LINE,
    ANSWER_PATTERN_WORD,
)

from ._metrics import accuracy, bootstrap_std
from ._pattern import pattern
from ._scorer import Scorer, scorer


class AnswerPattern(str, Enum):
    """Regular expressions for extracting answers from output.

    These expressions act on output prefixed with "ANSWER: ".
    """

    LETTER = ANSWER_PATTERN_LETTER
    """Extracts a single letter (used with multiple choice)."""

    WORD = ANSWER_PATTERN_WORD
    """Extracts one or more word characters (used for yes/no output)."""

    LINE = ANSWER_PATTERN_LINE
    """Extracts the rest of the line after ANSWER: (used for more complex output).

    Note that when using a LINE pattern your prompt should instruct the
    model to answer with a separate line at the end.
    """


@scorer(metrics=[accuracy(), bootstrap_std()])
def answer(type: Literal["letter", "word", "line"]) -> Scorer:
    """Scorer for model output that preceded answers with ANSWER:.

    Some solvers including multiple_choice solicit answers from
    the model prefaced with "ANSWER:". This scorer extracts
    answers of this form for comparison with the target.

    Note that you must specify a `type` for the answer scorer.

    Args:
      type: (Literal["letter", "word", "line"]): Type of answer
        to extract. "letter" is used with multiple choice and
        extracts a single letter; "word" will extract the next
        word (often used for yes/no answers); "line" will take
        the rest of the line (used for more more complex answers
        that may have embedded spaces). Note that when using
        "line" your prompt should instruct the model to answer
        with a separate line at the end.

    """
    match type:
        case "letter":
            return pattern(AnswerPattern.LETTER)
        case "word":
            return pattern(AnswerPattern.WORD)
        case "line":
            return pattern(AnswerPattern.LINE)



import click
from typing_extensions import Unpack

from inspect_ai._util.constants import DEFAULT_SERVER_HOST, DEFAULT_VIEW_PORT
from inspect_ai._view.view import view

from .common import CommonOptions, common_options, resolve_common_options


@click.command("view")
@click.option(
    "--recursive",
    type=bool,
    is_flag=True,
    default=True,
    help="Include all logs in log_dir recursively.",
)
@click.option(
    "--host",
    default=DEFAULT_SERVER_HOST,
    help="Tcp/Ip host",
)
@click.option("--port", default=DEFAULT_VIEW_PORT, help="TCP/IP port")
@common_options
def view_command(
    recursive: bool,
    host: str,
    port: int,
    **kwargs: Unpack[CommonOptions],
) -> None:
    """View evaluation logs."""
    # read common options
    (log_dir, log_level) = resolve_common_options(kwargs)

    # run the viewer
    view(
        log_dir=log_dir, recursive=recursive, host=host, port=port, log_level=log_level
    )



from ._file import (
    EvalLogInfo,
    list_eval_logs,
    read_eval_log,
    write_eval_log,
)
from ._log import (
    EvalConfig,
    EvalDataset,
    EvalError,
    EvalLog,
    EvalMetric,
    EvalPlan,
    EvalPlanStep,
    EvalResults,
    EvalRevision,
    EvalSample,
    EvalScorer,
    EvalSpec,
    EvalStats,
    LoggingLevel,
    LoggingMessage,
)

__all__ = [
    "EvalConfig",
    "EvalError",
    "EvalDataset",
    "EvalLog",
    "EvalMetric",
    "EvalPlan",
    "EvalPlanStep",
    "EvalResults",
    "EvalRevision",
    "EvalSample",
    "EvalScorer",
    "EvalSpec",
    "EvalStats",
    "EvalLogInfo",
    "LoggingLevel",
    "LoggingMessage",
    "list_eval_logs",
    "read_eval_log",
    "write_eval_log",
]



import os

from utils import (
    skip_if_no_anthropic,
    skip_if_no_google,
    skip_if_no_openai,
)

from inspect_ai import Task, eval, task
from inspect_ai.dataset import json_dataset
from inspect_ai.scorer import match
from inspect_ai.solver import generate, system_message

SYSTEM_MESSAGE = """
For the following exercise, it is important that you answer with only a single word or numeric value in brackets. For example, [22] or [house]. Do not include any discussion, narrative, or rationale, just a single value in brackets.
"""


@task
def images():
    return Task(
        dataset=json_dataset(os.path.join("tests", "test_images", "images.jsonl")),
        plan=[system_message(SYSTEM_MESSAGE), generate()],
        scorer=match(),
    )


def check_images(model):
    eval(images, model)


@skip_if_no_google
def test_google_images():
    check_images("google/gemini-pro-vision")


@skip_if_no_openai
def test_openai_images():
    check_images("opeanai/gpt-4")


@skip_if_no_anthropic
def test_anthropic_images():
    check_images("anthropic/claude-3-sonnet-20240229")



import os

from openai.types.chat import (
    ChatCompletion,
)
from typing_extensions import override

from inspect_ai._util.constants import DEFAULT_MAX_TOKENS
from inspect_ai.model._providers.util import model_base_url

from .._model import (
    ChatCompletionChoice,
    GenerateConfig,
    Logprob,
    Logprobs,
)
from .openai import (
    OpenAIAPI,
    chat_message_assistant,
)
from .util import as_stop_reason


def chat_choices_from_response_together(
    response: ChatCompletion,
) -> list[ChatCompletionChoice]:
    choices = list(response.choices)
    choices.sort(key=lambda c: c.index)
    logprobs_dicts = [
        choice.logprobs.model_dump() if choice.logprobs is not None else None
        for choice in choices
    ]
    logprobs_models: list[Logprobs | None] = []
    for logprob_dict in logprobs_dicts:
        if logprob_dict is None:
            logprobs_models.append(logprob_dict)
            continue
        tokens = logprob_dict["tokens"]
        token_logprobs = logprob_dict["token_logprobs"]
        logprobs_sequence = []
        for token, logprob in zip(tokens, token_logprobs):
            logprobs_sequence.append(
                Logprob(
                    token=token,
                    logprob=logprob,
                    bytes=list(map(ord, token)),
                    top_logprobs=None,
                )
            )
        logprobs_models.append(Logprobs(content=logprobs_sequence))
    return [
        ChatCompletionChoice(
            message=chat_message_assistant(choice.message),
            stop_reason=as_stop_reason(choice.finish_reason),
            logprobs=logprobs,
        )
        for choice, logprobs in zip(choices, logprobs_models)
    ]


class TogetherAIAPI(OpenAIAPI):
    def __init__(
        self,
        model_name: str,
        base_url: str | None = None,
        config: GenerateConfig = GenerateConfig(),
    ) -> None:
        api_key = os.environ.get("TOGETHER_API_KEY", None)
        if not api_key:
            raise RuntimeError("TOGETHER_API_KEY environment variable not set")
        base_url = model_base_url(base_url, "TOGETHER_BASE_URL")
        base_url = base_url if base_url else "https://api.together.xyz/v1"
        super().__init__(
            model_name=model_name, base_url=base_url, config=config, api_key=api_key
        )

    # Together uses a default of 512 so we bump it up
    @override
    def max_tokens(self) -> int:
        return DEFAULT_MAX_TOKENS

    # Together has a slightly different logprobs structure to OpenAI, so we need to remap it.
    def _chat_choices_from_response(
        self, response: ChatCompletion
    ) -> list[ChatCompletionChoice]:
        return chat_choices_from_response_together(response)



from random import random

from utils import skip_if_no_openai

from inspect_ai import Task, eval, eval_retry, task
from inspect_ai.dataset import Sample
from inspect_ai.scorer import match
from inspect_ai.solver import Generate, TaskState, generate, solver


@solver
def failing_solver():
    async def solve(state: TaskState, generate: Generate):
        if random() > 0.33:
            raise ValueError("Eval failed!")

        return state

    return solve


@task
def failing_task():
    return Task(
        dataset=[Sample(input="Say hello", target="hello")],
        plan=[failing_solver(), generate()],
        scorer=match(),
    )


@skip_if_no_openai
def test_eval_retry():
    # run eval with a solver that fails 2/3 times
    failing_eval = f"{__file__}@failing_task"
    log = eval(failing_eval, limit=1)[0]

    # note the task id so we can be certain it remains the same
    task_id = log.eval.task_id

    # retry until we succeed (confirming the task_id is stable)
    while log.status != "success":
        log = eval_retry(log)[0]
        assert log.eval.task_id == task_id



import json
import os
from typing import Any, cast

from openai import APIStatusError, AsyncAzureOpenAI, AsyncOpenAI, RateLimitError
from openai._types import NOT_GIVEN
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionAssistantMessageParam,
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartParam,
    ChatCompletionContentPartTextParam,
    ChatCompletionMessage,
    ChatCompletionMessageParam,
    ChatCompletionMessageToolCallParam,
    ChatCompletionNamedToolChoiceParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolMessageParam,
    ChatCompletionToolParam,
    ChatCompletionUserMessageParam,
)
from openai.types.shared_params.function_definition import FunctionDefinition
from typing_extensions import override

from inspect_ai._util.constants import DEFAULT_MAX_RETRIES
from inspect_ai._util.images import image_as_data_uri
from inspect_ai._util.url import is_data_uri, is_http_url

from .._model import (
    ChatCompletionChoice,
    ChatMessage,
    ChatMessageAssistant,
    Content,
    GenerateConfig,
    Logprobs,
    ModelAPI,
    ModelOutput,
    ModelUsage,
)
from .._tool import ToolCall, ToolChoice, ToolFunction, ToolInfo
from .._util import chat_api_tool
from .util import as_stop_reason, model_base_url

OPENAI_API_KEY = "OPENAI_API_KEY"
AZURE_OPENAI_API_KEY = "AZURE_OPENAI_API_KEY"
AZUREAI_OPENAI_API_KEY = "AZUREAI_OPENAI_API_KEY"


class OpenAIAPI(ModelAPI):
    def __init__(
        self,
        model_name: str,
        base_url: str | None = None,
        config: GenerateConfig = GenerateConfig(),
        api_key: str | None = None,
        **model_args: Any,
    ) -> None:
        # call super
        super().__init__(model_name=model_name, base_url=base_url, config=config)

        # resolve api_key
        is_azure = False
        if not api_key:
            api_key = os.environ.get(
                AZUREAI_OPENAI_API_KEY, os.environ.get(AZURE_OPENAI_API_KEY, None)
            )
            if api_key:
                is_azure = True
            else:
                api_key = os.environ.get(OPENAI_API_KEY, None)
                if not api_key:
                    raise ValueError(
                        f"No {OPENAI_API_KEY} or {AZUREAI_OPENAI_API_KEY} found."
                    )

        # save api_key for connection_key
        self.api_key = api_key

        # azure client
        if is_azure:
            # resolve base_url
            base_url = model_base_url(
                base_url,
                [
                    "AZUREAI_OPENAI_BASE_URL",
                    "AZURE_OPENAI_BASE_URL",
                    "AZURE_OPENAI_ENDPOINT",
                ],
            )
            if not base_url:
                raise ValueError(
                    "You must provide a base URL when using OpenAI on Azure. Use the AZUREAI_OPENAI_BASE_URL "
                    + " environment variable or the --model-base-url CLI flag to set the base URL."
                )

            self.client: AsyncAzureOpenAI | AsyncOpenAI = AsyncAzureOpenAI(
                api_key=api_key,
                azure_endpoint=base_url,
                azure_deployment=model_name,
                max_retries=(
                    config.max_retries if config.max_retries else DEFAULT_MAX_RETRIES
                ),
                **model_args,
            )
        else:
            self.client = AsyncOpenAI(
                api_key=api_key,
                base_url=model_base_url(base_url, "OPENAI_BASE_URL"),
                max_retries=(
                    config.max_retries if config.max_retries else DEFAULT_MAX_RETRIES
                ),
                **model_args,
            )

    async def generate(
        self,
        input: list[ChatMessage],
        tools: list[ToolInfo],
        tool_choice: ToolChoice,
        config: GenerateConfig,
    ) -> ModelOutput:
        # resolve max tokens (ignore type check so NotGiven is valid)
        config.max_tokens = config.max_tokens if config.max_tokens else NOT_GIVEN  # type: ignore
        # unlike text models, vision models require a max_tokens (and set it to a very low
        # default, see https://community.openai.com/t/gpt-4-vision-preview-finish-details/475911/10)
        OPENAI_IMAGE_DEFAULT_TOKENS = 4096
        if "vision" in self.model_name:
            if isinstance(config.max_tokens, int):
                config.max_tokens = max(config.max_tokens, OPENAI_IMAGE_DEFAULT_TOKENS)
            else:
                config.max_tokens = OPENAI_IMAGE_DEFAULT_TOKENS

        # normalize to openai messages
        messages = await as_openai_chat_messages(input)
        try:
            # generate completion
            response: ChatCompletion = await self.client.chat.completions.create(
                messages=messages,
                tools=chat_tools(tools) if len(tools) > 0 else NOT_GIVEN,
                tool_choice=(
                    chat_tool_choice(tool_choice) if len(tools) > 0 else NOT_GIVEN
                ),
                **self.completion_params(config),
            )
            choices = self._chat_choices_from_response(response)
            return ModelOutput(
                model=response.model,
                choices=choices,
                usage=(
                    ModelUsage(
                        input_tokens=response.usage.prompt_tokens,
                        output_tokens=response.usage.completion_tokens,
                        total_tokens=response.usage.total_tokens,
                    )
                    if response.usage
                    else None
                ),
            )
        except APIStatusError as e:
            completion, error = handle_content_filter_error(e)
            return ModelOutput.from_content(
                model=self.model_name,
                content=completion,
                stop_reason="content_filter",
                error=str(error) if error else None,
            )

    def _chat_choices_from_response(
        self, response: ChatCompletion
    ) -> list[ChatCompletionChoice]:
        # adding this as a method so we can override from other classes (e.g together)
        return chat_choices_from_response(response)

    @override
    def is_rate_limit(self, ex: BaseException) -> bool:
        if isinstance(ex, RateLimitError):
            # Do not retry on these rate limit errors
            if (
                "Request too large" not in ex.message
                and "You exceeded your current quota" not in ex.message
            ):
                return True
        return False

    @override
    def connection_key(self) -> str:
        """Scope for enforcing max_connections (could also use endpoint)."""
        return self.api_key

    def completion_params(self, config: GenerateConfig) -> dict[str, Any]:
        return dict(
            model=self.model_name,
            stream=False,  # Code below assumes this is not a streaming response
            frequency_penalty=(
                config.frequency_penalty
                if config.frequency_penalty is not None
                else NOT_GIVEN
            ),
            stop=config.stop_seqs if config.stop_seqs is not None else NOT_GIVEN,
            max_tokens=config.max_tokens,
            presence_penalty=(
                config.presence_penalty
                if config.presence_penalty is not None
                else NOT_GIVEN
            ),
            logit_bias=config.logit_bias if config.logit_bias else NOT_GIVEN,
            seed=config.seed if config.seed is not None else NOT_GIVEN,
            temperature=(
                config.temperature
                if config.temperature is not None
                else (
                    1  # TogetherAPI requires temperature w/ num_choices
                    if config.num_choices is not None
                    else NOT_GIVEN
                )
            ),
            top_p=config.top_p if config.top_p is not None else NOT_GIVEN,
            timeout=(
                float(config.timeout) if config.timeout is not None else NOT_GIVEN
            ),
            n=config.num_choices if config.num_choices is not None else NOT_GIVEN,
            logprobs=config.logprobs if config.logprobs is not None else NOT_GIVEN,
            top_logprobs=(
                config.top_logprobs if config.top_logprobs is not None else NOT_GIVEN
            ),
        )


async def as_openai_chat_messages(
    messages: list[ChatMessage],
) -> list[ChatCompletionMessageParam]:
    return [await openai_chat_message(message) for message in messages]


async def openai_chat_message(message: ChatMessage) -> ChatCompletionMessageParam:
    if message.role == "system":
        return ChatCompletionSystemMessageParam(role=message.role, content=message.text)
    elif message.role == "user":
        return ChatCompletionUserMessageParam(
            role=message.role,
            content=(
                message.content
                if isinstance(message.content, str)
                else [
                    await as_chat_completion_part(content)
                    for content in message.content
                ]
            ),
        )
    elif message.role == "assistant":
        if message.tool_calls:
            return ChatCompletionAssistantMessageParam(
                role=message.role,
                content=message.text,
                tool_calls=[chat_tool_call(call) for call in message.tool_calls],
            )
        else:
            return ChatCompletionAssistantMessageParam(
                role=message.role, content=message.text
            )
    elif message.role == "tool":
        return ChatCompletionToolMessageParam(
            role=message.role,
            content=(
                f"Error: {message.tool_error}" if message.tool_error else message.text
            ),
            tool_call_id=str(message.tool_call_id),
        )
    else:
        raise ValueError(f"Unexpected message role {message.role}")


def chat_tool_call(tool_call: ToolCall) -> ChatCompletionMessageToolCallParam:
    return ChatCompletionMessageToolCallParam(
        id=tool_call.id,
        function=dict(
            name=tool_call.function, arguments=json.dumps(tool_call.arguments)
        ),
        type=tool_call.type,
    )


def chat_tools(tools: list[ToolInfo]) -> list[ChatCompletionToolParam]:
    chat_tools = [chat_api_tool(tool) for tool in tools]
    return [
        ChatCompletionToolParam(
            type=tool["type"], function=cast(FunctionDefinition, tool["function"])
        )
        for tool in chat_tools
    ]


def chat_tool_choice(tool_choice: ToolChoice) -> ChatCompletionToolChoiceOptionParam:
    if isinstance(tool_choice, ToolFunction):
        return ChatCompletionNamedToolChoiceParam(
            type="function", function=dict(name=tool_choice.name)
        )
    else:
        return tool_choice


def chat_tool_calls(message: ChatCompletionMessage) -> list[ToolCall] | None:
    if message.tool_calls:
        return [
            ToolCall(
                id=call.id,
                function=call.function.name,
                arguments=json.loads(call.function.arguments),
                type="function",
            )
            for call in message.tool_calls
        ]
    else:
        return None


def chat_choices_from_response(response: ChatCompletion) -> list[ChatCompletionChoice]:
    choices = list(response.choices)
    choices.sort(key=lambda c: c.index)
    return [
        ChatCompletionChoice(
            message=chat_message_assistant(choice.message),
            stop_reason=as_stop_reason(choice.finish_reason),
            logprobs=(
                Logprobs(**choice.logprobs.model_dump())
                if choice.logprobs is not None
                else None
            ),
        )
        for choice in choices
    ]


def chat_message_assistant(message: ChatCompletionMessage) -> ChatMessageAssistant:
    return ChatMessageAssistant(
        content=message.content or "",
        source="generate",
        tool_calls=chat_tool_calls(message),
    )


async def as_chat_completion_part(
    content: Content,
) -> ChatCompletionContentPartParam:
    if content.type == "text":
        return ChatCompletionContentPartTextParam(type="text", text=content.text)
    else:
        # API takes URL or base64 encoded file. If it's a remote file or
        # data URL leave it alone, otherwise encode it
        image_url, detail = (
            (content.image, "auto")
            if isinstance(content.image, str)
            else (content.image, content.detail)
        )

        if not is_http_url(image_url) and not is_data_uri(image_url):
            image_url = await image_as_data_uri(image_url)

        return ChatCompletionContentPartImageParam(
            type="image_url",
            image_url=dict(url=image_url, detail=cast(Any, detail)),
        )


# Azure throws an APIStatusError (w/ status 400) when its content
# moderation policies are triggered, which invalidates the entire
# eval run with an error. In this case we'd rather not end the run
# entirely but rather return the error as the model "message" and
# then record the error in ModelOutput metadata. Note that OpenAI
# does not exhibit this behavior (it just returns the completion
# "Sorry, but I can't assist with that."
def handle_content_filter_error(e: APIStatusError) -> tuple[str, object | None]:
    CANT_ASSIST = "Sorry, but I can't assist with that."
    if e.status_code == 400:
        if isinstance(e.body, dict) and "message" in e.body.keys():
            message = str(e.body.get("message"))
            return message, e.body
        else:
            return CANT_ASSIST, e.body
    else:
        raise e



import os
from typing import Any

from .._dataset import (
    Dataset,
    FieldSpec,
    RecordToSample,
)
from .csv import csv_dataset
from .json import json_dataset


def file_dataset(
    file: str,
    sample_fields: FieldSpec | RecordToSample | None = None,
    dialect: str = "unix",
    encoding: str = "utf-8",
    name: str | None = None,
    fs_options: dict[str, Any] = {},
) -> Dataset:
    """Dataset read from a JSON or CSV file.

    The `file_dataset` function supports reading from CSV and JSON files
    (and automatically delegates to the appropriate function to do so)

    Args:
        file (str): Path to JSON or CSV file. Can be a local filesystem path or
            a path to an S3 bucket (e.g. "s3://my-bucket"). Use `fs_options`
            to pass arguments through to the `S3FileSystem` constructor.
        sample_fields (SampleFieldSpec | RecordToSample): Method of mapping underlying
            fields in the data source to Sample objects. Pass `None` if the data is already
            stored in `Sample` form (i.e. has "input" and "target" columns.); Pass a
            `SampleFieldSpec` to specify mapping fields by name; Pass a `RecordToSample` to
            handle mapping with a custom function.
        dialect (str): CSV dialect ("unix" or "excel", defaults to "unix"). Only
            applies to reading CSV files.
        encoding (str): Text encoding for file (defaults to "utf-8").
        name (str): Optional name for dataset (for logging). If not specified,
            defaults to the stem of the filename
        fs_options (dict[str, Any]): Optional. Addional arguments to pass through
            to the filesystem provider (e.g. `S3FileSystem`). Use `{"anon": True }`
            if you are accessing a public S3 bucket with no credentials.

    Returns:
        Dataset read from JSON or CSV file.
    """
    ext = os.path.splitext(file)[1].lower()

    match ext:
        case ".json" | ".jsonl":
            return json_dataset(
                json_file=file,
                sample_fields=sample_fields,
                encoding=encoding,
                name=name,
                fs_options=fs_options,
            )
        case ".csv":
            return csv_dataset(
                csv_file=file,
                sample_fields=sample_fields,
                dialect=dialect,
                encoding=encoding,
                name=name,
                fs_options=fs_options,
            )
        case _:
            raise ValueError(f"No dataset reader for file with extension {ext}")



from importlib.metadata import version


def pip_dependency_error(feature: str, dependencies: list[str]) -> Exception:
    return ModuleNotFoundError(
        f"ERROR: {feature} requires optional dependencies. "
        f"Install with:\n\npip install {' '.join(dependencies)}\n"
    )


def module_version_error(
    feature: str, package: str, required_version: str
) -> Exception:
    return ModuleNotFoundError(
        f"ERROR: {feature} requires at least version {required_version} of package {package} "
        f"(you have version {version(package)} installed).\n\n"
        f"Upgrade with:\n\npip install --upgrade {package}\n"
    )


def exception_message(ex: BaseException) -> str:
    return getattr(ex, "message", repr(ex))



import os
from typing import Type, TypeVar

import pytest

from inspect_ai.dataset import (
    Dataset,
    FieldSpec,
    Sample,
    csv_dataset,
    example_dataset,
    file_dataset,
    json_dataset,
)

T_ds = TypeVar("T_ds")

# test functions are parameterized by dataset type and input file
csv = (csv_dataset, "samples.csv")
json = (json_dataset, "samples.json")
jsonl = (file_dataset, "samples.jsonl")
dataset_params = [csv, json, jsonl]


# test reading a dataset using default configuration
@pytest.mark.parametrize("type,file", dataset_params)
def test_dataset(type: Type[T_ds], file: str) -> None:
    dataset: Dataset = type.__call__(dataset_path(file))
    assert_sample(dataset[0])


# test reading a dataset with an explcit fields specification
@pytest.mark.parametrize("type,file", dataset_params)
def test_dataset_fields(type: Type[T_ds], file: str) -> None:
    dataset: Dataset = type.__call__(
        dataset_path(file), sample_fields=sample_field_spec
    )
    assert_sample(dataset[0])


# test reading a dataset with a custom data_to_sample function
@pytest.mark.parametrize("type,file", dataset_params)
def test_dataset_fields_fn(type: Type[T_ds], file: str) -> None:
    dataset: Dataset = type.__call__(
        dataset_path(file),
        sample_fields=data_to_sample,
    )
    assert_sample(dataset[0])


def test_dataset_read_id() -> None:
    dataset = example_dataset(
        "biology_qa",
        FieldSpec(input="question", target="answer", id="id"),
    )
    assert dataset[0].id == "q1"


sample_field_spec = FieldSpec(input="input", target="label", metadata=["extra"])


def data_to_sample(data: dict) -> Sample:
    return Sample(
        input=str(data.get("input")),
        target=str(data.get("label")),
        metadata={"extra": data.get("extra")},
    )


def assert_sample(sample: Sample) -> None:
    assert sample.input == "Say 'Hello, World'"
    assert sample.target == "Hello, World"
    if sample.metadata:
        assert sample.metadata.get("extra") == "data"


def dataset_path(file: str) -> str:
    return os.path.join("tests", "test_dataset", file)


def example_path(*paths: str) -> str:
    return os.path.join("examples", "/".join(paths))



from inspect_ai import Task, task
from inspect_ai.dataset import FieldSpec, example_dataset
from inspect_ai.scorer import model_graded_qa
from inspect_ai.solver import generate, use_tools, web_search


@task
def biology_qa() -> Task:
    return Task(
        dataset=example_dataset(
            name="biology_qa",
            sample_fields=FieldSpec(
                input="question",
                target="answer"
            ),
        ),
        plan=[use_tools(web_search()), generate()],
        scorer=model_graded_qa(),
    )




import abc
import random
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterator,
    Optional,
    Sequence,
    Union,
    overload,
)

from pydantic import BaseModel, Field
from typing_extensions import override

from inspect_ai.model import ChatMessage

if TYPE_CHECKING:
    from _typeshed import SupportsRichComparison


class Sample(BaseModel):
    r"""Sample to be used in an evaluation task.

    Args:
        input (str | list[ChatMessage]): The input to be submitted to the model.
        choices (list[str] | None): Optional. List of available answer choices
           (used only for multiple-choice evals).
        target (str | list[str] | None): Optional. Ideal target output. May be a literal value
            or narrative text to be used by a model grader.
        id (int | str | None): Optional. Unique identifier for sample.
        metadata (dict | None): Optional. Arbitrary metadata associated with the sample.
    """

    input: str | list[ChatMessage]
    """The input to be submitted to the model."""

    choices: list[str] | None = Field(default=None)
    """List of available answer choices (used only for multiple-choice evals)."""

    target: str | list[str] = Field(default="")
    """Ideal target output. May be a literal value or narrative text to be used by a model grader."""

    id: int | str | None = Field(default=None)
    """Unique identifier for sample."""

    metadata: dict[str, Any] | None = Field(default=None)
    """Arbitrary metadata associated with the sample."""


def sample_input_len(sample: Sample) -> int:
    """Measures the length of a samples `input` field.

    The default length function use in `Dataset.sort()`.

    Args:
        sample (Sample): A Sample to be used in an evaluation task.
    """
    return (
        len(sample.input)
        if isinstance(sample.input, str)
        else sum(len(inp.text) for inp in sample.input)
    )


DatasetRecord = dict[str, Any]

DatasetReader = Iterator[DatasetRecord]


class Dataset(Sequence[Sample], abc.ABC):
    r"""A sequence of Sample objects.

    Datasets provide sequential access (via conventional indexes or slicing)
    to a collection of Sample objects.
    """

    @abc.abstractproperty
    def name(self) -> str | None: ...

    @abc.abstractproperty
    def location(self) -> str | None: ...

    @overload
    def __getitem__(self, index: int) -> Sample: ...

    @overload
    def __getitem__(self, index: slice) -> "Dataset": ...

    @abc.abstractmethod
    def __getitem__(self, index: Union[int, slice]) -> Union[Sample, "Dataset"]: ...

    @abc.abstractmethod
    def __len__(self) -> int: ...

    @abc.abstractmethod
    def shuffle(self, seed: int | None = None) -> None:
        """Shuffle the order of the dataset (in place).

        Args:
           seed: (int | None): Random seed for shuffling (optional).
        """

    @abc.abstractmethod
    def sort(
        self,
        reverse: bool = False,
        key: Optional[Callable[[Sample], "SupportsRichComparison"]] = sample_input_len,
    ) -> None:
        """Sort the dataset (in place) in ascending order and return None.

        If a key function is given, apply it once to each list item and sort them, ascending or descending, according to their function values.

        The key function defaults to measuring the length of the sample's input field.

        Args:
            reverse (bool): if true, sort in descending order. Defaults to False.
            key (Callable[[Any], Any]): a callable mapping each item to a numeric value (optional, defaults to sample_input_len).
        """

    @abc.abstractmethod
    def filter(
        self, predicate: Callable[[Sample], bool], name: str | None = None
    ) -> "Dataset":
        """Filter the dataset using a predicate.

        Args:
          predicate (Callable[[Sample], bool]): Filtering function.
          name (str | None): Name for filtered dataset (optional).

        Returns:
          Filtered dataset.
        """


class FieldSpec(BaseModel):
    r"""Specification for mapping data source fields to sample fields.

    Args:
        input (str): Name of the field containing the sample input.
        target (str): Name of the field containing the sample target.
        choices (str): Optional. Name of field containing the list of answer choices.
        id (str): Optional. Unique identifier for the sample.
        metadata (list[str] | None): List of additional field names that should be read as metadata.
    """

    input: str = Field(default="input")
    """Name of the field containing the sample input."""

    target: str = Field(default="target")
    """Name of the field containing the sample target."""

    choices: str = Field(default="choices")
    """Name of field containing the list of answer choices."""

    id: str = Field(default="id")
    """ Unique identifier for the sample."""

    metadata: list[str] | None = Field(default=None)
    """List of additional field names that should be read as metadata."""


RecordToSample = Callable[[DatasetRecord], Sample]
r"""Callable that maps raw dictionary record to a Sample."""


class MemoryDataset(Dataset):
    r"""A Dataset stored in memory."""

    def __init__(
        self,
        samples: list[Sample],
        name: str | None = None,
        location: str | None = None,
    ) -> None:
        r"""A dataset of samples held in an in-memory list.

        Datasets provide sequential access (via conventional indexes or slicing)
        to a collection of Sample objects. The ListDataset is explicitly
        initialized with a list that is held in memory.

        Args:
            samples (list[Sample]): The list of sample objects.
            name (str | None): Optional name for dataset.
            location (str | None): Optional location for dataset.
        """
        self.samples = samples
        self._name = name
        self._location = location

    @override
    @property
    def name(self) -> str | None:
        """Dataset name."""
        return self._name

    @override
    @property
    def location(self) -> str | None:
        """Dataset location."""
        return self._location

    @overload
    def __getitem__(self, index: int) -> Sample: ...

    @overload
    def __getitem__(self, index: slice) -> Dataset: ...

    @override
    def __getitem__(self, index: Union[int, slice]) -> Union[Sample, Dataset]:
        if isinstance(index, int):
            return self.samples[index]
        else:
            return MemoryDataset(
                samples=self.samples[index], name=self.name, location=self.location
            )

    @override
    def __len__(self) -> int:
        return len(self.samples)

    @override
    def shuffle(self, seed: int | None = None) -> None:
        if seed:
            random.Random(seed).shuffle(self.samples)
        else:
            random.shuffle(self.samples)

    @override
    def sort(
        self,
        reverse: bool = False,
        key: Optional[Callable[[Sample], "SupportsRichComparison"]] = sample_input_len,
    ) -> None:
        self.samples.sort(reverse=reverse, key=key)

    @override
    def filter(
        self, predicate: Callable[[Sample], bool], name: str | None = None
    ) -> "MemoryDataset":
        return MemoryDataset(
            name=name or self.name,
            location=self.location,
            samples=[sample for sample in self if predicate(sample)],
        )



import pytest
from utils import skip_if_no_openai

from inspect_ai import Task, eval_async
from inspect_ai._util.registry import registry_info
from inspect_ai.dataset import Sample
from inspect_ai.solver import (
    Generate,
    Plan,
    TaskState,
    chain_of_thought,
    generate,
    plan,
    solver,
)


@plan(fancy=True)
def my_plan() -> Plan:
    return Plan(steps=[chain_of_thought(), generate()])


@skip_if_no_openai
@pytest.mark.asyncio
async def test_plan_cleanup():
    @solver
    def failing_solver():
        async def solve(state: TaskState, generate: Generate):
            raise ValueError("Eval failed!")

        return solve

    cleaned_up = False

    def cleanup(state):
        nonlocal cleaned_up
        cleaned_up = True

    task = Task(
        dataset=[Sample(input="Say hello.", target="Hello")],
        plan=Plan(
            steps=[chain_of_thought(), failing_solver(), generate()], cleanup=cleanup
        ),
    )

    result = await eval_async(task, model="openai/gpt-4")

    assert result[0].status == "error"
    assert cleaned_up


def test_plan_registration():
    plan = my_plan()
    assert registry_info(plan).name == "my_plan"


def test_plan_attribs():
    plan = my_plan()
    assert registry_info(plan).metadata["attribs"]["fancy"] is True



"""
Measuring Massive Multitask Language Understanding

Dan Hendrycks, Collin Burns, Steven Basart, Andy Zou,
Mantas Mazeika, Dawn Song, Jacob Steinhardt
https://arxiv.org/abs/2009.03300

Based on: https://github.com/openai/simple-evals/blob/main/mmlu_eval.py

# eval all subjects w/ 500 randomly selected samples
inspect eval mmlu.py@mmlu --limit 500

# add chain of thought
inspect eval mmlu.py@mmlu --limit 500 -T cot=true

# eval selected subjects
inspect eval mmlu.py@mmlu -T subjects=anatomy
inspect eval mmlu.py@mmlu -T subjects=astronomy
inspect eval mmlu.py@mmlu -T subjects=anatomy,astronomy
"""

from inspect_ai import Task, task
from inspect_ai.dataset import Sample, csv_dataset
from inspect_ai.model import GenerateConfig
from inspect_ai.scorer import answer
from inspect_ai.solver import multiple_choice


# map records to inspect sample
def record_to_sample(record):
    return Sample(
        input=record["Question"],
        choices=[
            str(record["A"]),
            str(record["B"]),
            str(record["C"]),
            str(record["D"]),
        ],
        target=record["Answer"],
        metadata={"subject": record["Subject"]},
    )


# read dataset globally so it can be shared by all of the tasks
# (shuffle so that --limit draws from multiple subjects)
dataset = csv_dataset(
    csv_file="datasets/mmlu.csv",
    sample_fields=record_to_sample,
    shuffle=True,
)


@task
def mmlu(subjects=[], cot=False):
    # filter dataset if requested
    subjects = subjects if isinstance(subjects, list) else [subjects]
    if len(subjects) > 0:
        task_dataset = dataset.filter(
            name=f"{dataset.name}-{'-'.join(subjects)}",
            predicate=lambda sample: sample.metadata["subject"] in subjects,
        )
    else:
        task_dataset = dataset

    # return task
    return Task(
        dataset=task_dataset,
        plan=multiple_choice(cot=cot),
        scorer=answer("letter"),
        config=GenerateConfig(temperature=0.5),
    )



"""
Training Verifiers to Solve Math Word Problems

Karl Cobbe, Vineet Kosaraju, Mohammad Bavarian, Mark Chen, Heewoo Jun, Lukasz Kaiser, Matthias Plappert, Jerry Tworek, Jacob Hilton, Reiichiro Nakano, Christopher Hesse, John Schulman
https://arxiv.org/abs/2110.14168

# run with default fewshots (10)
inspect eval gsm8k.py

# run with less  or no fewshots
inspect eval gsm8k.py -T fewshot=5
inspect eval gsm8k.py -T fewshot=false
"""

from inspect_ai import Task, task
from inspect_ai.dataset import Sample, hf_dataset
from inspect_ai.scorer import match
from inspect_ai.solver import generate, prompt_template, system_message


def record_to_sample(record):
    DELIM = "####"
    input = record["question"]
    answer = record["answer"].split(DELIM)
    target = answer.pop().strip()
    reasoning = DELIM.join(answer)
    return Sample(input=input, target=target, metadata={"reasoning": reasoning.strip()})


def sample_to_fewshot(sample):
    return (
        f"{sample.input}\n\nReasoning:\n"
        + f"{sample.metadata['reasoning']}\n\n"
        + f"ANSWER: {sample.target}"
    )


# setup for problem + instructions for providing answer
MATH_PROMPT_TEMPLATE = """
Solve the following math problem step by step. The last line of your response should be of the form "ANSWER: $ANSWER" (without quotes) where $ANSWER is the answer to the problem.

{prompt}

Remember to put your answer on its own line at the end in the form "ANSWER: $ANSWER" (without quotes) where $ANSWER is the answer to the problem, and you do not need to use a \\boxed command.

Reasoning:
""".strip()


@task
def gsm8k(fewshot=10, fewshot_seed=42):
    # build plan dynamically (may or may not be doing fewshot)
    plan = [prompt_template(MATH_PROMPT_TEMPLATE), generate()]
    if fewshot:
        fewshots = hf_dataset(
            path="gsm8k",
            data_dir="main",
            split="train",
            sample_fields=record_to_sample,
            shuffle=True,
            seed=fewshot_seed,
            limit=fewshot,
        )
        plan.insert(
            0,
            system_message(
                "\n\n".join([sample_to_fewshot(sample) for sample in fewshots])
            ),
        )

    # define task
    return Task(
        dataset=hf_dataset(
            path="gsm8k",
            data_dir="main",
            split="test",
            sample_fields=record_to_sample,
        ),
        plan=plan,
        scorer=match(numeric=True),
    )
